#!/usr/bin/env bash
import argparse
import os
import sys


def build(args):
    path = os.path.join(os.path.dirname(__file__), "src")
    gcc = "gcc -fPIC -shared -o"

    cmd_list = []
    if args.buf:
        cmd_list.append("{} {} buf.c buf.h".format(gcc, os.path.join(os.path.pardir, "buf.so")))

    for cmd in cmd_list:
        cmd = "cd {}; {}".format(path, cmd)
        print("# {}".format(cmd))

        if os.system(cmd) >> 8:
            raise RuntimeError('{} failed!'.format(cmd))

    print("Succeed.")


if __name__ == '__main__':
    PARSER = argparse.ArgumentParser(description='Build')

    PARSER.add_argument('-b', '--buf', action='store_true', help='Build buffer')

    ARGS = PARSER.parse_args()

    sys.exit(build(ARGS))



#!/usr/bin/env python
"""
Created on 2017/3/3

@author: yyang
"""
import os
import platform
import re
import random
from ctypes import POINTER, cast, sizeof, addressof, memmove, memset, pointer
from ctypes import c_uint8, c_uint16, c_uint32, c_uint64, string_at, CDLL, c_void_p

from tool.device.nvme.ctype import Ctype
from utils import log


class NVMeMalloc(object):
    def __init__(self, length=4096):
        filename = 'buf.dll' if platform.system() == 'Windows' else 'buf.so'
        path = os.path.realpath(os.path.join(os.getcwd(), "lib", "driver", "dll", filename))
        self._dll = CDLL(path)
        self._dll.nvme_alloc.restype = c_uint64

        self._ptr = self._dll.nvme_alloc(length)
        assert self._ptr != 0, 'Malloc {:#x} Failed!'.format(length)

    def __del__(self):
        self.free()

    def mem_copy(self, dptr, size):
        self._dll.nvme_copy(self.ptr, dptr, size)

    def free(self):
        if self._ptr:
            self._dll.nvme_free(self.ptr)

    @property
    def ptr(self):
        return c_void_p(self._ptr)

    @property
    def address(self):
        return self._ptr


# noinspection PyTypeChecker
class Malloc(object):
    def __init__(self, length=1, types=c_uint8):
        filename = 'buf.dll' if platform.system() == 'Windows' else 'buf.so'
        path = os.path.realpath(os.path.join(os.getcwd(), "lib", "driver", "dll", filename))
        self.m_dll = CDLL(path)

        self._types = types  # types of item
        self._len = length  # count of items
        self._sizeof = sizeof(self._types)  # size of item
        self._size = self._len * self._sizeof  # count of bytes
        # noinspection PyTypeChecker,PyCallingNonCallable
        self._buf = (self._types * self._len)()  # buffer instance

        self._lp_uint8 = cast(self._buf, POINTER(c_uint8))
        self._lp_uint16 = cast(self._buf, POINTER(c_uint16))
        self._lp_uint32 = cast(self._buf, POINTER(c_uint32))
        self._lp_uint64 = cast(self._buf, POINTER(c_uint64))

    def __getitem__(self, key):
        return self._buf[key]

    def __setitem__(self, key, val):
        self._buf[key] = val

    def realloc(self, types=c_uint8):
        assert self._size % sizeof(types) == 0, 'Buffer is unaligned as type of {}'.format(types)

        self._types = types
        self._sizeof = sizeof(self._types)
        self._len = self._size // self._sizeof
        _buf = (self._types * self._len)()
        memmove(_buf, self._buf, self._size)
        self._buf = _buf

        self._lp_uint8 = cast(self._buf, POINTER(c_uint8))
        self._lp_uint16 = cast(self._buf, POINTER(c_uint16))
        self._lp_uint32 = cast(self._buf, POINTER(c_uint32))
        self._lp_uint64 = cast(self._buf, POINTER(c_uint64))

    @property
    def len(self):
        return self._len

    @property
    def size(self):
        return self._size

    @property
    def sizeof(self):
        return self._sizeof

    @property
    def types(self):
        return self._types

    @property
    def buf(self):
        return self._buf

    @property
    def address(self):
        return addressof(self._buf)

    @property
    def pointer(self):
        return pointer(self._buf)

    def cast(self, types):
        return cast(self._buf, POINTER(types))

    def convert(self, types, index=0):
        return cast(self._buf, POINTER(types))[index]

    def memset(self, byte):
        memset(self._buf, byte, self.size)

    def memmove(self, offset, source, start, length):
        memmove(self.address + offset, source.address + start, length)

    def memcopy(self, stream, offset=0, length=0):
        data = [ord(i) for i in list(stream)]
        size = min(length, len(data), self.size) if length > 0 else min(len(data), self.size)
        for i in range(size):
            self.set_uint8(offset + i, data[i])

    def get_uint8(self, offset):
        return self._lp_uint8[offset]

    def get_uint16(self, offset):
        return self._lp_uint16[offset // 2]

    def get_uint32(self, offset):
        return self._lp_uint32[offset // 4]

    def get_uint64(self, offset):
        return self._lp_uint64[offset // 8]

    def set_uint8(self, offset, value):
        self._lp_uint8[offset] = value

    def set_uint16(self, offset, value):
        self._lp_uint16[offset // 2] = value

    def set_uint32(self, offset, value):
        self._lp_uint32[offset // 4] = value

    def set_uint64(self, offset, value):
        self._lp_uint64[offset // 8] = value

    def translate(self, start, size=0, types=c_uint8):
        _sizeof = sizeof(types)

        assert start >= 0 and size >= 0, 'Start: {:#x} or Size: {:#x} < 0!'.format(start, size)
        assert start + size <= self.size, 'Start: {:#x} + Size: {:#x} > Max: {:#x}!'.format(start, size, self.size)
        assert start % _sizeof == 0, 'Start: {:#x} is not aligned with type: {}!'.format(start, types)
        assert size % _sizeof == 0, 'Size: {:#x} is not aligned with type: {}!'.format(size, types)

        length = size // _sizeof if size > 0 else (self.size - start) // _sizeof
        offset = start // _sizeof
        return cast(self._buf, POINTER(types)), offset, length

    def fill_uint32_seq(self, value=0, start=0, size=0, step=1):
        self.m_dll.fill_uint32_seq(*self.translate(start, size, c_uint32), value, step)

    def fill_uint32_fix(self, value=0, start=0, size=0):
        self.m_dll.fill_uint32_fix(*self.translate(start, size, c_uint32), value)

    def fill_uint32_rnd(self, seed=0, start=0, size=0):
        self.m_dll.fill_uint32_rnd(*self.translate(start, size, c_uint32), seed)

    def compare_uint32_seq(self, value=0, start=0, size=0, step=1):
        return self.m_dll.compare_uint32_seq(*self.translate(start, size, c_uint32), value, step)

    def compare_uint32_fix(self, value=0, start=0, size=0):
        return self.m_dll.compare_uint32_fix(*self.translate(start, size, c_uint32), value)

    def compare_uint32_rnd(self, seed=0, start=0, size=0):
        return self.m_dll.compare_uint32_rnd(*self.translate(start, size, c_uint32), seed)

    def diff_uint32_seq(self, value=0, start=0, size=0):
        buffer, offset, length = self.translate(start, size, c_uint32)
        log.DUMP("  Address  :  Expected   Actual")
        log.DUMP("-----------------------------------")
        for i in range(offset, offset + length):
            if buffer[i] != value + i:
                log.DUMP("{:#010x} : {:#010x} {:#010x}".format(i * 4, value + i, buffer[i]))
        log.DUMP("-----------------------------------")

    def diff_uint32_fix(self, value=0, start=0, size=0):
        buffer, offset, length = self.translate(start, size, c_uint32)
        log.DUMP("  Address  :  Expected   Actual")
        log.DUMP("-----------------------------------")
        for i in range(offset, offset + length):
            if buffer[i] != value:
                log.DUMP("{:#010x} : {:#010x} {:#010x}".format(i * 4, value, buffer[i]))
        log.DUMP("-----------------------------------")

    def diff_uint32_rnd(self, seed=0, start=0, size=0):
        _random = random.Random(seed)
        buffer, offset, length = self.translate(start, size, c_uint32)
        log.DUMP("  Address  :  Expected   Actual")
        log.DUMP("-----------------------------------")
        for i in range(offset, offset + length):
            excepted = _random.randrange(0, 1 << 32)
            if buffer[i] != excepted:
                log.DUMP("{:#010x} : {:#010x} {:#010x}".format(i * 4, excepted, buffer[i]))
        log.DUMP("-----------------------------------")

    def fill_uint8_fix(self, value=0, start=0, size=0):
        self.m_dll.fill_uint8_fix(*self.translate(start, size, c_uint8), value)

    def look_idx_uint32(self, value):
        return self.m_dll.look_idx_uint32(self.buf, self.size // 4, value)

    def compare(self, dst, src_start=0, src_size=0, dst_start=None, dst_size=None):
        dst_start = src_start if dst_start is None else dst_start
        dst_size = src_size if dst_size is None else dst_size
        src_size = min(src_size, self.size - src_start) if src_size > 0 else self.size - src_start
        dst_size = min(dst_size, dst.size - dst_start) if dst_size > 0 else dst.size - dst_start

        return self.m_dll.compare(self.buf, dst.buf, src_start, dst_start, min(src_size, dst_size))

    def dump(self, skip=None, no_rsv=True, only_values=False):
        Ctype(self._buf).dump(skip=skip, no_rsv=no_rsv, only_values=only_values)

    def dump_diff(self, dst, src_start=0, src_size=0, dst_start=None, dst_size=None):
        dst_start = dst_start if dst_start is not None else src_start
        dst_size = dst_size if dst_size is not None else src_size

        src_buf, src_off, src_len = self.translate(src_start, src_size, c_uint32)
        dst_buf, dst_off, dst_len = dst.translate(dst_start, dst_size, c_uint32)
        log.DUMP("")

        log.DUMP("  ADDR_A  /  ADDR_B   :   DATA_A     DATA_B")
        log.DUMP("---------------------------------------------")
        for i in range(min(src_len, dst_len)):
            src_index = i + src_off
            dst_index = i + dst_off
            if src_buf[src_index] != dst_buf[dst_index]:
                log.DUMP("{:#010x}/{:#010x} : {:#010x} {:#010x}".format(
                    src_index * 4, dst_index * 4, src_buf[src_index], dst_buf[dst_index]))

    def dump_uint8(self, start=0, size=0):
        log.DUMP("          0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f")
        log.DUMP("---------------------------------------------------------")

        buffer, offset, length = self.translate(start, size, c_uint8)

        for i in range(offset, offset + length, 0x10):
            data = ["{:02x}".format(d) for d in buffer[i:min(i + 0x10, offset + length)]]
            log.DUMP("{:#08x}: {}".format(i, " ".join(data)))

    def dump_uint32(self, start=0, size=0):
        log.DUMP("         0        1        2        3       ")
        log.DUMP("--------------------------------------------")

        buffer, offset, length = self.translate(start, size, c_uint32)

        for i in range(offset, offset + length, 0x4):
            data = ["{:08x}".format(d) for d in buffer[i:min(i + 0x4, offset + length)]]
            log.DUMP("{:#08x}: {}".format(i * 4, " ".join(data)))

    def dump_uint8_uart(self, start=0, size=0):
        buffer, offset, length = self.translate(start, size, c_uint8)

        for i in range(offset, offset + length, 0x10):
            data = ["{:02x}".format(d) for d in buffer[i:min(i + 0x10, offset + length)]]
            log.DUMP("{}".format(" ".join(data)))

    def set_sub_buffer(self, offset, size, types=c_uint8):
        p_buff = cast(self._buf, POINTER(c_uint8))
        d_list = Malloc(size//sizeof(types), types)
        if (offset + size) > self._len:
            raise RuntimeError("[ERR] out of range(offset: {}, size: {})".format(offset, size))
        for off in range(offset, offset + size):
            d_list.set_uint8(off - offset, p_buff[off])
        return d_list

    def mask(self, buff):
        for i in range(buff.size):
            mask = buff.get_uint8(i)
            if mask == 1:
                self.set_uint8(i, 0)

    def get_string(self, offset, length):
        dat_list = list()
        p_buff = cast(self._buf, POINTER(c_uint8))
        for i in range(length):
            off = offset + i
            ret = "{:#04x}".format(p_buff[off])
            dat_list.append(ret)
        dat_list.reverse()
        ret = "".join(dat_list)
        return ret

    @staticmethod
    def to_number(array):
        string = '{{:0{}x}}'.format(sizeof(array) // len(array) * 2)
        data = [string.format(i) for i in array]
        return int(''.join(data), 16)

    @staticmethod
    def to_string(ctype, length=0):
        length = length if length else sizeof(ctype)
        string = string_at(addressof(ctype), length).decode('utf-8', errors='ignore')
        return re.sub(r'[\x00-\x1f]', '', string).strip()

    def set_multi_bytes(self, offset, length, buff):
        for i in range(length):
            self.set_uint8(offset + i, buff.get_uint8(i))

    def set_multi_bytes_fix(self, offset, length, val):
        for i in range(length):
            self.set_uint8(offset + i, val)

    def set_multi_dword(self, stream, offset=0, length=1):
        for i in range(length):
            self.set_uint32(offset + i * 4, stream[i])

    def set_multi_qword(self, stream, offset=0, length=1):
        for i in range(length):
            self.set_uint64(offset + i * 8, stream[i])

    def write(self, path, mode='wb'):
        with open(path, mode) as fd:
            fd.write(self._buf)

    def read(self, path):
        with open(path, "rb") as fd:
            memmove(self._buf, fd.read(), self._size)

    def write_string(self, path, mode='wb'):
        with open(path, mode) as fd:
            string = "".join([chr(i) for i in cast(self._buf, POINTER(c_uint8))])
            fd.write(string)


def generate_big_number(array):
    """
    Generate as a big number with data of array
    :param array: Ctypes array
    :return:      Unsigned number
    """
    total = 0
    bits = sizeof(type(array)) // len(array) * 8
    for i, num in enumerate(array):
        total += (num << (i * bits))
    return total


def get_string(array, little_endian=True):
    """
    Generate as a hex number with data of array
    :param array:           Ctypes array
    :param little_endian:   Little endian or big endian
    :return:                String of hex number without 0x
    """
    content_list = ["{:02x}".format(x) for x in array]
    if little_endian:
        content_list.reverse()
    return "".join(content_list)



#!/usr/bin/env python
# pylint: disable=superfluous-parens,missing-docstring,too-few-public-methods
"""
Created on Apr 19, 2019

@author: yyang
"""
from ctypes import Structure, c_uint8, c_uint16, c_uint32, c_uint64
from tool.device.nvme.ctype import Uint128, Byte


class PowerStateDescriptorDataStructure(Structure):
    """Identify - Power State Descriptor Data Structure"""
    _pack_ = 1
    _fields_ = [
        ('MP', c_uint16),  # Maximum Power (MP)
        ('RSV0', c_uint8),
        ('MXPS', c_uint8, 1),  # Max Power Scale (MXPS)
        ('NOPS', c_uint8, 1),  # Non-Operational State (NOPS)
        ('RSV1', c_uint8, 6),
        ('ENLAT', c_uint32),  # Entry Latency (ENLAT)
        ('EXLAT', c_uint32),  # Exit Latency (EXLAT)
        ('RRT', c_uint8, 5),  # Relative Read Throughput (RRT)
        ('RSV2', c_uint8, 3),
        ('RRL', c_uint8, 5),  # Relative Read Latency (RRL)
        ('RSV3', c_uint8, 3),
        ('RWT', c_uint8, 5),  # Relative Write Throughput (RWT)
        ('RSV4', c_uint8, 3),
        ('RWL', c_uint8, 5),  # Relative Write Latency (RWL)
        ('RSV5', c_uint8, 3),
        ('IDLP', c_uint16),  # Idle Power (IDLP)
        ('RSV6', c_uint8, 6),
        ('IPS', c_uint8, 2),  # Idle Power Scale (IPS)
        ('RSV7', c_uint8),
        ('ACTP', c_uint16),  # Active Power (ACTP)
        ('APW', c_uint8, 3),  # Active Power Workload (APW)
        ('RSV8', c_uint8, 3),
        ('APS', c_uint8, 2),  # Active Power Scale (APS)
        ('RSV9', c_uint8 * 9),
    ]
    desc = {
        'MP': 'Maximum Power', 'MXPS': 'Max Power Scale', 'NOPS': 'Non-Operational State',
        'ENLAT': 'Entry Latency', 'EXLAT': 'Exit Latency', 'RRT': 'Relative Read Throughput',
        'RRL': 'Relative Read Latency', 'RWT': 'Relative Write Throughput', 'RWL': 'Relative Write Latency',
        'IDLP': 'Idle Power', 'IPS': 'Idle Power Scale', 'ACTP': 'Active Power', 'APW': 'Active Power Workload',
        'APS': 'Active Power Scale'
    }


class ControllerDataStructure(Structure):
    """Identify - Identify Controller Data Structure"""
    _pack_ = 1
    _fields_ = [
        ('VID', c_uint16),  # PCI Vendor ID
        ('SSVID', c_uint16),  # PCI Subsystem Vendor ID
        ('SN', Byte * 20),  # Serial number
        ('MN', Byte * 40),  # Model number for the NVM subsystem
        ('FR', Byte * 8),  # Firmware revision for the NVM subsystem
        ('RAB', c_uint8),  # Recommended Arbitration Burst size
        ('IEEE', c_uint32, 24),  # IEEE OUI Identifier
        ('CMIC', c_uint32, 8),  # Controller Multi-Path I/O and Namespace Sharing Capabilities
        ('MDTS', c_uint8),  # Maximum Data Transfer Size
        ('CNTLID', c_uint16),  # Controller ID
        ('VER', c_uint32),  # Version
        ('RTD3R', c_uint32),  # RTD3 Resume Latency
        ('RTD3E', c_uint32),  # RTD3 Entry Latency
        ('OAES', c_uint32),  # Optional Asynchronous Events Supported
        ('CTRATT', c_uint32),  # Controller Attributes (CTRATT)
        ('RSV0', c_uint8 * 12),
        ('FGUID', Byte * 12),  # FRU Globally Unique Identifier (FGUID)
        ('RSV1', c_uint8 * 116),
        ('MMIS', Byte * 16),  # NVMe Management Interface Specification
        ('OACS', c_uint16),  # Optional Admin Command Support
        ('ACL', c_uint8),  # Abort Command Limit
        ('AERL', c_uint8),  # Asynchronous Event Request Limit
        ('FRMW', c_uint8),  # Firmware Updates
        ('LPA', c_uint8),  # Log Page Attributes
        ('ELPE', c_uint8),  # Error Log Page Entries
        ('NPSS', c_uint8),  # Number of Power States Support
        ('AVSCC', c_uint8),  # Admin Vendor Specific Command Configuration
        ('APSTA', c_uint8),  # Autonomous Power State Transition Attributes
        ('WCTEMP', c_uint16),  # Warning Composite Temperature Threshold
        ('CCTEMP', c_uint16),  # Critical Composite Temperature Threshold
        ('MTFA', c_uint16),  # Maximum Time for Firmware Activation
        ('HMPRE', c_uint32),  # Host Memory Buffer Preferred Size
        ('HMMIN', c_uint32),  # Host Memory Buffer Minimum Size
        ('TNVMCAP', Uint128),  # Total NVM Capacity
        ('UNVMCAP', Uint128),  # Unallocated NVM Capacity
        ('RPMBS', c_uint32),  # Replay Protected Memory Block Support
        ('EDSTT', c_uint16),  # Extended Device Self-test Time (EDSTT)
        ('DSTO', c_uint8),  # Device Self-test Options (DSTO)
        ('FWUG', c_uint8),  # Firmware Update Granularity (FWUG)
        ('KAS', c_uint16),  # Keep Alive Support (KAS)
        ('HCTMA', c_uint16),  # Host Controlled Thermal Management Attributes (HCTMA)
        ('MNTMT', c_uint16),  # Minimum Thermal Management Temperature (MNTMT)
        ('MXTMT', c_uint16),  # Maximum Thermal Management Temperature (MXTMT)
        ('SANICAP', c_uint32),  # Sanitize Capabilities (SANICAP)
        ('RSV2', c_uint8 * 180),
        ('SQES', c_uint8),  # Submission Queue Entry Size
        ('CQES', c_uint8),  # Completion Queue Entry Size
        ('MAXCMD', c_uint16),  # Maximum Outstanding Commands (MAXCMD)
        ('NN', c_uint32),  # Number of Namespaces
        ('ONCS', c_uint16),  # Optional NVM Command Support
        ('FUSES', c_uint16),  # Fused Operation Support
        ('FNA', c_uint8),  # Format NVM Attributes
        ('VWC', c_uint8),  # Volatile Write Cache
        ('AWUN', c_uint16),  # Atomic Write Unit Normal
        ('AWUPF', c_uint16),  # Atomic Write Unit Power Fail
        ('NVSCC', c_uint8),  # NVM Vendor Specific Command Configuration
        ('RSV3', c_uint8),
        ('ACWU', c_uint16),  # Atomic Compare & Write Unit
        ('RSV4', c_uint8 * 2),
        ('SGLS', c_uint32),  # SGL Support
        ('RSV5', c_uint8 * 28),
        ('SUBNQN', Byte * 256),  # NVM Subsystem NVMe Qualified Name
        ('RSV6', c_uint8 * 768),
        ('PSD', PowerStateDescriptorDataStructure * 32),  # Power State Descriptor
        ('VS', Byte * 1024),  # Vendor Specific
    ]
    desc = {
        'VID': 'PCI Vendor ID', 'SSVID': 'PCI Subsystem Vendor ID', 'SN': 'Serial number',
        'MN': 'Model number for the NVM subsystem', 'FR': 'Firmware revision for the NVM subsystem',
        'RAB': 'Recommended Arbitration Burst size', 'IEEE': 'IEEE OUI Identifier',
        'CMIC': 'Controller Multi-Path I/O and Namespace Sharing Capabilities', 'MDTS': 'Maximum Data Transfer Size',
        'CNTLID': 'Controller ID', 'VER': 'Version', 'RTD3R': 'RTD3 Resume Latency', 'RTD3E': 'RTD3 Entry Latency',
        'OAES': 'Optional Asynchronous Events Supported', 'CTRATT': 'Controller Attributes',
        'FGUID': 'FRU Globally Unique Identifier', 'MMIS': 'NVMe Management Interface Specification',
        'OACS': 'Optional Admin Command Support', 'ACL': 'Abort Command Limit',
        'AERL': 'Asynchronous Event Request Limit', 'FRMW': 'Firmware Updates', 'LPA': 'Log Page Attributes',
        'ELPE': 'Error Log Page Entries', 'NPSS': 'Number of Power States Support',
        'AVSCC': 'Admin Vendor Specific Command Configuration', 'APSTA': 'Autonomous Power State Transition Attributes',
        'WCTEMP': 'Warning Composite Temperature Threshold', 'CCTEMP': 'Critical Composite Temperature Threshold',
        'MTFA': 'Maximum Time for Firmware Activation', 'HMPRE': 'Host Memory Buffer Preferred Size',
        'HMMIN': 'Host Memory Buffer Minimum Size', 'TNVMCAP': 'Total NVM Capacity',
        'UNVMCAP': 'Unallocated NVM Capacity', 'RPMBS': 'Replay Protected Memory Block Support',
        'EDSTT': 'Extended Device Self-test Time', 'DSTO': 'Device Self-test Options',
        'FWUG': 'Firmware Update Granularity', 'KAS': 'Keep Alive Support',
        'HCTMA': 'Host Controlled Thermal Management Attributes', 'MNTMT': 'Minimum Thermal Management Temperature',
        'MXTMT': 'Maximum Thermal Management Temperature', 'SANICAP': 'Sanitize Capabilities',
        'SQES': 'Submission Queue Entry Size', 'CQES': 'Completion Queue Entry Size',
        'MAXCMD': 'Maximum Outstanding Commands', 'NN': 'Number of Namespaces',
        'ONCS': 'Optional NVM Command Support', 'FUSES': 'Fused Operation Support',
        'FNA': 'Format NVM Attributes', 'VWC': 'Volatile Write Cache', 'AWUN': 'Atomic Write Unit Normal',
        'AWUPF': 'Atomic Write Unit Power Fail', 'NVSCC': 'NVM Vendor Specific Command Configuration',
        'ACWU': 'Atomic Compare & Write Unit', 'SGLS': 'SGL Support', 'SUBNQN': 'NVM Subsystem NVMe Qualified Name',
        'PSD': 'Power State Descriptor',
        'VS': 'Vendor Specific'
    }


class LBAFormatDataStructure(Structure):
    """Identify - LBA Format Data Structure, NVM Command Set Specific"""
    _pack_ = 1
    _fields_ = [
        ('MS', c_uint32, 16),  # Metadata Size
        ('LBADS', c_uint32, 8),  # LBA Data Size
        ('RP', c_uint32, 2),  # Relative Performance
        ('RSV', c_uint32, 6),
    ]
    desc = {
        'MS': 'Metadata Size', 'LBADS': 'LBA Data Size', 'RP': 'Relative Performance'
    }


class NamespaceDataStructure(Structure):
    """Identify - Identify Namespace Data Structure, NVM Command Set Specific"""
    _pack_ = 1
    _fields_ = [
        ('NSIZE', c_uint64),  # Namespace Size
        ('NCAP', c_uint64),  # Namespace Capacity
        ('NUSE', c_uint64),  # Namespace Utilization
        ('NSFEAT', c_uint8),  # Namespace Features
        ('NLBAF', c_uint8),  # Number of LBA Formats
        ('FLBAS', c_uint8),  # Formatted LBA Size
        ('MC', c_uint8),  # Metadata Capabilities
        ('DPC', c_uint8),  # End-to-end Data Protection Capabilities
        ('DPS', c_uint8),  # End-to-end Data Protection Type Settings
        ('NMIC', c_uint8),  # Namespace Multi-path I/O and Namespace Sharing Capabilities
        ('RESCAP', c_uint8),  # Reservation Capabilities
        ('FPI', c_uint8),  # Format Progress Indicator
        ('DLFEAT', c_uint8),  # Deallocate Logical Block Features (DLFEAT)
        ('NAWUN', c_uint16),  # Namespace Atomic Write Unit Normal (NAWUN)
        ('NAWUPF', c_uint16),  # Namespace Atomic Write Unit Power Fail (NAWUPF)
        ('NACWU', c_uint16),  # Namespace Atomic Compare & Write Unit
        ('NABSN', c_uint16),  # Namespace Atomic Boundary Size Normal
        ('NABO', c_uint16),  # Namespace Atomic Boundary Offset
        ('NABSPF', c_uint16),  # Namespace Atomic Boundary Size Power Fail
        ('NOIOB', c_uint16),  # Namespace Optimal IO Boundary (NOIOB)
        ('NVMCAP', Uint128),  # NVM Capacity (NVMCAP)
        ('RSV0', c_uint8 * 40),
        ('NGUID', Byte * 16),  # Namespace Globally Unique Identifier
        ('EUI64', c_uint64),  # IEEE Extended Unique Identifier
        ('LBAF', LBAFormatDataStructure * 16),  # LBA Format Support
        ('RSV1', c_uint8 * 192),
        ('VS', Byte * 3712),  # Vendor Specific
    ]
    desc = {
        'NSIZE': 'Namespace Size', 'NCAP': 'Namespace Capacity', 'NUSE': 'Namespace Utilization',
        'NSFEAT': 'Namespace Features', 'NLBAF': 'Number of LBA Formats', 'FLBAS': 'Formatted LBA Size',
        'MC': 'Metadata Capabilities', 'DPC': 'End-to-end Data Protection Capabilities',
        'DPS': 'End-to-end Data Protection Type Settings',
        'NMIC': 'Namespace Multi-path I/O and Namespace Sharing Capabilities',
        'RESCAP': 'Reservation Capabilities', 'FPI': 'Format Progress Indicator',
        'DLFEAT': 'Deallocate Logical Block Features', 'NAWUN': 'Namespace Atomic Write Unit Normal',
        'NAWUPF': 'Namespace Atomic Write Unit Power Fail', 'NACWU': 'Namespace Atomic Compare & Write Unit',
        'NABSN': 'Namespace Atomic Boundary Size Normal', 'NABO': 'Namespace Atomic Boundary Offset',
        'NABSPF': 'Namespace Atomic Boundary Size Power Fail', 'NOIOB': 'Namespace Optimal IO Boundary',
        'NVMCAP': 'NVM Capacity', 'NGUID': 'Namespace Globally Unique Identifier',
        'EUI64': 'IEEE Extended Unique Identifier', 'LBAF': 'LBA Format Support', 'VS': 'Vendor Specific'
    }


class PrimaryControllerCap(Structure):
    """Identify - PrimaryController Data Structure"""
    _pack_ = 1
    _fields_ = [
        ('CNTLID', c_uint16),  # Controller Identifier
        ('PORTID', c_uint16),  # Port Identifier
        ('CRT', c_uint8),  # Controller Resource Types
        ('RSV0', c_uint8 * 27),
        ('VQFRT', c_uint32),  # VQ Resources Flexible Total
        ('VQRFA', c_uint32),  # VQ Resources Flexible Assigned
        ('VQRFAP', c_uint16),  # VQ Resources Flexible Allocated to Primary
        ('VQPRT', c_uint16),  # VQ Resources Private Total
        ('VQFRSM', c_uint16),  # VQ Resources Flexible Secondary Maximum
        ('VQGRAN', c_uint16),  # VQ Flexible Resource Preferred Granularity
        ('RSV1', c_uint8 * 16),
        ('VIFRT', c_uint32),  # VI Resources Flexible Total
        ('VIRFA', c_uint32),  # VI Resources Flexible Assigned
        ('VIRFAP', c_uint16),  # VI Resources Flexible Allocated to Primary
        ('VIPRT', c_uint16),  # VI Resources Private Total
        ('VIFRSM', c_uint16),  # VI Resources Flexible Secondary Maximum
        ('VIGRAN', c_uint16),  # VI Flexible Resource Preferred Granularity
        ('RSV2', c_uint8 * 4016),  # NVM Set Attributes Entry
    ]
    desc = {
        'CNTLID': 'Controller Identifier', 'PORTID': 'Port Identifier', 'CRT': 'Controller Resource Types',
        'VQFRT': 'VQ Resources Flexible Total', 'VQRFA': 'VQ Resources Flexible Assigned',
        'VQRFAP': 'VQ Resources Flexible Allocated to Primary', 'VQPRT': 'VQ Resources Private Total',
        'VQFRSM': 'VQ Resources Flexible Secondary Maximum', 'VQGRAN': 'VQ Flexible Resource Preferred Granularity',
        'VIFRT': 'VI Resources Flexible Total', 'VIFRA': 'VI Resources Flexible Assigned',
        'VIRFAP': 'VI Resources Flexible Allocated to Primary', 'VIPRT': 'VI Resources Private Total',
        'VIFRSM': 'I Resources Flexible Secondary Maximum', 'VIGRAN': 'VI Flexible Resource Preferred Granularity'
    }


class SecondaryControllerListEntry(Structure):
    """Identify - SecondaryControllerListEntry Data Structure"""
    _pack_ = 1
    _fields_ = [
        ('SCID', c_uint16),  # Secondary Controller Identifier
        ('PCID', c_uint16),  # Primary Controller Identifier
        ('SCS', c_uint8),  # Secondary Controller State
        ('RSV0', c_uint8 * 3),
        ('VFN', c_uint16),  # Virtual Function Number
        ('NVQ', c_uint16),  # Number of VQ Flexible Resources Assigned
        ('NVI', c_uint16),  # Number of VI Flexible Resources Assigned
        ('RSV1', c_uint8 * 18),
    ]
    desc = {
        'SCID': 'Secondary Controller Identifier', 'PCID': 'Primary Controller Identifier',
        'SCS': 'Secondary Controller State', 'VFN': 'Virtual Function Number',
        'NVQ': 'Number of VQ Flexible Resources Assigned', 'NVI': 'Number of VI Flexible Resources Assigned'
    }


class SecondaryControllerList(Structure):
    """Identify - SecondaryControllerList Data Structure"""
    _pack_ = 1
    _fields_ = [
        ('NI', c_uint8),  # Number of Identifiers
        ('RSV0', c_uint8 * 31),
        ('ENTRY', SecondaryControllerListEntry * 127),  # Secondary Controller List
    ]
    desc = {
        'NI': 'Number of Identifiers', 'ENTRY': 'Secondary Controller List',
    }


class NamespaceList(Structure):
    _pack_ = 1
    _fields_ = [
        ('ID', c_uint32 * 1024),  # Namespace ID
    ]
    desc = {
        'ID': 'Namespace ID'
    }


class ControllerList(Structure):
    _pack_ = 1
    _fields_ = [
        ('NUM', c_uint16),  # Number of Identifiers
        ('ID', c_uint16 * 2047),  # Controller ID
    ]
    desc = {
        'NUM': 'Number of Identifiers', 'ID': 'Controller ID'
    }


IDENTIFY_CNS = {
    0x0: NamespaceDataStructure,
    0x1: ControllerDataStructure,
    0x2: NamespaceList,
    0x10: NamespaceList,
    0x11: NamespaceDataStructure,
    0x12: ControllerList,
    0x13: ControllerList,
    0x14: PrimaryControllerCap,
    0x15: SecondaryControllerList,
}




import subprocess
import os
import time
from utils.system import decorate_exception


class SerialDownloader(object):

    def __init__(self):
        self.root_path = os.getcwd()
        self.script_path = os.path.join(self.root_path, "Utility")
        self.logs_path = os.path.join(self.root_path, "Logs", "Two_step")
        self.orig_log_folders = list()
        self.latest_log_folders = list()
        self.log_file = None
        self.log_path = None
        self.init_log_path()

    def init_log_path(self):
        if os.path.exists(self.logs_path) is False:
            os.mkdir(self.logs_path)

    def get_orig_logs(self):
        log_dirs = os.listdir(self.logs_path)
        for item in log_dirs:
            if os.path.isfile(os.path.join(self.logs_path, item)):
                self.orig_log_folders.append(os.path.join(self.logs_path, item))

    def gen_cmd_line(self, com_port, fw_path, oakgate, pre_bin_path):
        command_line = "cd /d {} && python two_step_download.py --oakgate={} --firmwarePath={} --preBinPath={} --serialPort={}"\
            .format(self.script_path, oakgate, fw_path, pre_bin_path, com_port)
        return command_line

    def execute_command(self, cmd):
        ret = os.system(cmd)
        unused_err = ""
        return ret, unused_err

    def get_bin_path(self, fw_path, vol, commit):
        if os.path.isfile(fw_path):
            bin_path = fw_path
        else:
            bin_path = self.get_fw_path(fw_path, vol, commit)
        return bin_path

    def get_spi_bin(self, fw_path):
        spi_bin_path = None
        if os.path.isdir(fw_path):
            spi_bin_path = self.get_spi_path(fw_path)
        return spi_bin_path

    def get_fw_path(self, fw_path, vol, commit):
        for file_name in os.listdir(fw_path):
            if os.path.isfile(os.path.join(fw_path, file_name)):
                if "_{}_".format(vol) in file_name and commit in file_name and file_name.endswith(".cap"):
                    return os.path.join(fw_path, file_name)
        return None

    def get_spi_path(self, fw_path):
        for file_name in os.listdir(fw_path):
            if os.path.isfile(os.path.join(fw_path, file_name)):
                if "spi" in file_name.lower() and file_name.endswith(".cap"):
                    return os.path.join(fw_path, file_name)
        return None

    def get_new_log(self):
        self.latest_log_folders = os.listdir(self.logs_path)
        new_logs = list()
        for item in self.latest_log_folders:
            log_item = os.path.join(self.logs_path, item)
            if os.path.isfile(log_item):
                if log_item not in self.orig_log_folders:
                    new_logs.append(log_item)
        return new_logs


    def create_error_log(self, content):
        error_log = os.path.join(self.logs_path, "error_{}.log".format(time.time()))
        err_file = open(error_log, "w")
        try:
            if type(content) is bytes:
                err_file.write(content.decode('utf-8', 'ignore'))
            else:
                err_file.write(str(content))
        except Exception as e:
            print("Write error failed", e)
        err_file.close()
        return error_log

    @decorate_exception
    def run(self, parameters):
        com_port = parameters["com"]
        fw_path = parameters["fw_path"]
        oakgate = parameters["ogt"]
        vol = parameters["vol"]
        commit = parameters["commit"]
        self.get_orig_logs()
        bin_path = self.get_bin_path(fw_path, vol, commit)
        pre_bin_path = self.get_spi_bin(fw_path)
        if bin_path is not None:
            command_line = self.gen_cmd_line(com_port, bin_path, oakgate, pre_bin_path)
            print(command_line)
            ret, unused_err = self.execute_command(command_line)
        else:
            ret = 1
            unused_err = ("Did not find fw bin at: {}".format(fw_path))
        error_log = self.create_error_log(unused_err)
        logs = self.get_new_log()
        logs.append(error_log)
        return ret, logs



# coding=utf-8
import time
import os
from multiprocessing import Queue
from utils import log
from utils.process import MyProcess
from test_framework.test_engine import NoseEngine, OakgateEngine
from utils.system import decorate_exception


class Runner(object):

    def __init__(self):
        self.results = list()
        self.process_run_ = None
        self.working_path = os.environ["working_path"]
        self.engine = self.get_engine()
        # self.log_path = self.get_log_path()

    def get_engine(self):
        if "test-platform" in self.working_path:
            engine = OakgateEngine()
        else:
            engine = NoseEngine()
        return engine

    def get_results(self):
        return self.results

    def stop(self):
        print("test runner . stop")
        if self.process_run_ is not None:
            ret = self.process_run_.stop()
        else:
            ret = -1
        return ret

    @decorate_exception
    def process_run(self, test_case, test_case_path=None, parameters=None, loop=1, timeout=0):
        value = None
        for item in range(loop):
            log.INFO("Run test in loop: %s", item)
            start_time = time.time()
            current_time = start_time
            queue = Queue()
            self.process_run_ = MyProcess(target=self.engine.run, args=(test_case, test_case_path, parameters, queue,))
            self.process_run_.daemon = True
            self.process_run_.start()
            if timeout > 0:
                while current_time - start_time < timeout:
                    current_time = time.time()
                    time.sleep(5)
                self.process_run_.stop()
            else:
                self.process_run_.join()
            value = queue.get(True)
            self.results.append(value)
        return value



#!/usr/bin/env python
"""
Created on 2017/3/3

@author: yyang
"""
import os
import sys
import zlib
import platform
from ctypes import c_uint8, c_uint16, c_uint32, c_uint64, Array, Union, Structure, CDLL
from ctypes import POINTER, cast, sizeof, addressof, memmove, memset, pointer
from .system import ROOT_PATH


def get_string(content, little_endian=True):
    """Get string"""
    content_list = ["{:#04x}".format(x) for x in content]
    if little_endian:
        content_list.reverse()
    return "".join(content_list)


def dump_line(name, data, indent=0):
    """Dump line"""
    if data == "-":
        if name == "-":
            print("-" * 132)
        else:
            print("| {:<128} |".format(name))
    else:
        line = list()
        line.append("{:<15}".format("{}{}".format(" " * indent, name)))
        if data == "HEAD":
            line.append("{:<18}".format("HEX"))
            line.append("{:<22}".format("DEC"))
            line.append("{:<64}".format("BIN"))
        elif isinstance(data, (bool, int, float, type(sys.maxsize + 1))):
            line.append("{:<#18x}".format(data))
            line.append("{:<22}".format(data))
            line.append("{:<64}".format(bin(data)[2:]))
        else:
            line.append("{:<18}".format(data))
            line.append("{:<22}".format(data))
            line.append("{:<64}".format(data))

        print("| {} |".format(" | ".join(line)))


def dump(buf, indent, skip):
    """Dump UnionType/StructType to STDOUT"""
    if not isinstance(type(buf), (type(Union), type(Structure))):
        raise RuntimeError("Error type({})".format(type(buf)))

    skip = skip if isinstance(skip, list) else [skip]

    for field in getattr(buf, '_fields_'):
        name, types = field[0], field[1]
        for item in skip:
            if item in name and item:
                break
        else:
            value = getattr(buf, name)

            if isinstance(types, (type(Union), type(Structure))):
                dump_line(name, "", indent)
                dump(value, indent + 2, skip)
            elif isinstance(types, type(Array)):
                for i, item in enumerate(value):
                    if isinstance(type(item), (type(Union), type(Structure))):
                        dump_line("{}[{}]".format(name, i), "", indent)
                        dump(item, indent + 2, skip)
                    else:
                        dump_line("{}[{}]".format(name, i), item, indent)
            else:
                dump_line(name, value, indent)


class Malloc(object):
    """
    Malloc memory
    """

    def __init__(self, length=1, types=c_uint8):
        self.m_types = types  # types of item
        self.m_len = length  # count of items
        self.m_sizeof = sizeof(self.m_types)  # size of item
        self.m_size = self.m_len * self.m_sizeof  # count of bytes
        # noinspection PyTypeChecker,PyCallingNonCallable
        self.m_buf = (self.m_types * self.m_len)()  # buffer instance

    def __getitem__(self, key):
        return self.m_buf[key]

    def __setitem__(self, key, val):
        self.m_buf[key] = val

    def realloc(self, length=sys.maxsize, types=c_uint8):
        """Realloc memory to other types"""
        self.m_types = types
        self.m_len = min(length, self.m_size // sizeof(types))
        self.m_sizeof = sizeof(self.m_types)
        self.m_size = self.m_len * self.m_sizeof
        self.m_buf = cast(self.m_buf, POINTER(types))
        return self

    def len(self):
        """Get length"""
        return self.m_len

    def size(self):
        """Get total size"""
        return self.m_size

    def sizeof(self):
        return self.m_sizeof

    def types(self):
        """Get type"""
        return self.m_types

    def buffer(self):
        """Get instance"""
        return self.m_buf

    def addressof(self):
        """Get address of"""
        return addressof(self.m_buf)

    def pointer(self):
        """Get pointer"""
        return pointer(self.m_buf)

    def cast(self, types):
        """Cast buffer to type lp"""
        return cast(self.m_buf, POINTER(types))

    def convert(self, types, index=0):
        """Cast buffer to type lp and return buf[index]"""
        return cast(self.m_buf, POINTER(types))[index]

    def memset(self, value):
        """Set memory"""
        memset(self.m_buf, value, self.m_size)

    def memmove(self, offset, source, start, length):
        """Copy memory"""
        memmove(self.addressof() + offset, source.addressof() + start, length)

    def memcopy(self, stream, offset=0, length=sys.maxsize):
        """Set buffer by stream"""
        data = [i for i in list(stream)]
        size = min(length, len(data), self.m_size)
        for i in range(size):
            self.set_uint8(offset + i, data[i])

    def dump(self, offset=0, length=sys.maxsize, skip=""):
        """Dump buffer"""
        length = min(length, self.m_len)
        print("")
        if "ctypes" in str(self.m_types):
            self.dump_uint8(offset, length * self.m_sizeof)
        else:
            dump_line(str(self.m_types), "-")
            for i in range(offset, offset + length):
                print("")
                dump_line("BUFFER[{}]:".format(i), "HEAD")
                dump_line("-", "-")
                dump(self.m_buf[i], 0, skip)
                dump_line("-", "-")
        print("")

    def translate(self, start, size, types=c_uint8):
        length = min(self.m_size - (start // sizeof(types)) * sizeof(types), size) // sizeof(types)
        offset = start // sizeof(types)
        memory = cast(self.m_buf, POINTER(types))
        return memory, offset, length

    def dump_diff(self, dest, src_start=0, src_size=sys.maxsize, dst_start=None, dst_size=None):
        dst_start = dst_start if dst_start is not None else src_start
        dst_size = dst_size if dst_size is not None else src_size

        src_buf, src_off, src_len = self.translate(src_start, src_size, c_uint32)
        dst_buf, dst_off, dst_len = dest.translate(dst_start, dst_size, c_uint32)
        print("")

        print("          ADDR_A  /  ADDR_B   :   DATA_A     DATA_B")
        print("-----------------------------------------------------")
        for i in range(min(src_len, dst_len)):
            src_index = i + src_off
            dst_index = i + dst_off
            if src_buf[src_index] != dst_buf[dst_index]:
                print("ADDRESS {:#010x}/{:#010x} : {:#010x} {:#010x}".format(
                    src_index * 4, dst_index * 4, src_buf[src_index], dst_buf[dst_index]))

    def dump_uint8(self, start=0, size=sys.maxsize):
        """Dump with uint8"""
        print("          0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f")
        print("---------------------------------------------------------")

        memory, offset, length = self.translate(start, size, c_uint8)

        for i in range(offset, offset + length, 0x10):
            data = ["{:02x}".format(d) for d in memory[i:min(i + 0x10, offset + length)]]
            print("{:#08x}: {}".format(i, " ".join(data)))

    def dump_uint32(self, start=0, size=sys.maxsize):
        """Dump with uint32"""
        print("         0        1        2        3       ")
        print("--------------------------------------------")

        memory, offset, length = self.translate(start, size, c_uint32)

        for i in range(offset, offset + length, 0x4):
            data = ["{:08x}".format(d) for d in memory[i:min(i + 0x4, offset + length)]]
            print("{:#08x}: {}".format(i, " ".join(data)))

    def dump_uint8_uart(self, start=0, size=sys.maxsize):
        """Dump with uint32"""
        memory, offset, length = self.translate(start, size, c_uint8)

        for i in range(offset, offset + length, 0x10):
            data = ["{:02x}".format(d) for d in memory[i:min(i + 0x10, offset + length)]]
            print("{}".format(" ".join(data)))

    def set_sub_buffer(self, offset, size):
        """Set sub buffer"""
        p_buff = cast(self.m_buf, POINTER(c_uint8))
        d_list = Malloc(size)
        if (offset + size) > self.m_len:
            raise RuntimeError("[ERR] out of range(offset: {}, size: {})".format(offset, size))
        for off in range(offset, offset + size):
            d_list.set_uint8(off - offset, p_buff[off])
        return d_list

    def mask(self, buff):
        """Mask data"""
        for i in range(buff.size()):
            mask = buff.get_uint8(i)
            if mask == 1:
                self.set_uint8(i, 0)

    def get_string(self, offset, length):
        """Get string of buffer"""
        dat_list = list()
        p_buff = cast(self.m_buf, POINTER(c_uint8))
        for i in range(length):
            off = offset + i
            ret = "{:#04x}".format(p_buff[off])
            dat_list.append(ret)
        dat_list.reverse()
        ret = "".join(dat_list)
        return ret

    def ascii_to_string(self, offset, length):
        dat_list = list()
        p_buff = cast(self.m_buf, POINTER(c_uint8))
        for i in range(length):
            off = offset + i
            ret = "{}".format(chr(p_buff[off]))
            dat_list.append(ret)
        ret = "".join(dat_list)
        return ret

    def get(self, types, index):
        if not 0 <= index < self.m_size // sizeof(types):
            raise RuntimeError("[ERR] out of range(types: {}, index: {})".format(types, index))

        return cast(self.m_buf, POINTER(types))[index]

    def set(self, types, index, value):
        if not 0 <= index < self.m_size // sizeof(types):
            raise RuntimeError("[ERR] out of range(types: {}, index: {})".format(types, index))

        cast(self.m_buf, POINTER(types))[index] = value

    def get_uint8(self, offset):
        return self.get(c_uint8, offset)

    def get_uint16(self, offset):
        return self.get(c_uint16, offset // 2)

    def get_uint32(self, offset):
        return self.get(c_uint32, offset // 4)

    def get_uint64(self, offset):
        return self.get(c_uint64, offset // 8)

    def set_uint8(self, offset, value):
        self.set(c_uint8, offset, value)

    def set_uint16(self, offset, value):
        self.set(c_uint16, offset // 2, value)

    def set_uint32(self, offset, value):
        self.set(c_uint32, offset // 4, value)

    def set_uint64(self, offset, value):
        self.set(c_uint64, offset // 8, value)

    def set_multi_bytes(self, offset, length, buff):
        """Set multiple bytes"""
        for i in range(length):
            self.set_uint8(offset + i, buff.get_uint8(i))

    def set_multi_bytes_fix(self, offset, length, val):
        """Set multiple bytes with fixed data"""
        for i in range(length):
            self.set_uint8(offset + i, val)

    def set_multi_dword(self, stream, offset=0, length=1):
        """Set buffer with multi-dword"""
        # data = map(ord, list(stream))
        for i in range(length):
            self.set_uint32(offset + i * 4, stream[i])

    def set_multi_qword(self, stream, offset=0, length=1):
        """Set buffer with multi-qword"""
        # data = map(ord, list(stream))
        for i in range(length):
            self.set_uint64(offset + i * 8, stream[i])

    def write(self, path, offset=0, mode='wb'):
        """Write buffer to file"""
        with open(path, mode) as fpath:
            fpath.seek(offset, 0)
            fpath.write(self.m_buf)

    def read(self, path):
        """Read from file to buffer"""
        with open(path, "rb") as fpath:
            memmove(self.m_buf, fpath.read(), self.m_size)

    def write_string(self, path, offset=0, mode='wb'):
        """Write buffer to file with string"""
        with open(path, mode) as fpath:
            fpath.seek(offset, 2)
            string = "".join([chr(i) for i in list(self.m_buf)])
            fpath.write(string)

    def copy_value(self, value):
        """Copy value"""
        memmove(self.m_buf, value, self.m_size)

    def crc32(self):
        """Get crc32"""
        return zlib.crc32(self.m_buf) & 0xffffffff



# coding=utf-8
# pylint: disable=dangerous-default-value
import re
import time
import platform
import subprocess
from multiprocessing import Process
from utils import log


class MyProcess(Process):

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        super(MyProcess, self).__init__(group, target, name, args, kwargs)
        log.INFO("MyProcess init")

    def stop(self):
        print("MyProcess . stop")
        if "windows" in platform.system().lower():
            log.INFO("system is windows")
            ret = self._kill_process_windows()
        else:
            log.INFO("system is linux")
            self._kill_process_linux()
            ret = 0
        return ret

    def get_subproecess(self, parent_pid, pid_information):
        sub_pids = re.findall("\s+([0-9]+)\s+{}".format(parent_pid), pid_information.decode('utf-8'))
        if sub_pids:
            for pid_ in sub_pids:
                pids_ = self.get_subproecess(pid_, pid_information)
                for item in pids_:
                    if item not in sub_pids:
                        sub_pids.append(item)
        return sub_pids

    def _kill_process_linux(self):
        cmd = "ps -ef"
        child2 = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        out = child2.communicate()
        sub_process_ids = self.get_subproecess(self.pid, out[0])
        sub_process_ids.append(self.pid)
        log.INFO(sub_process_ids)
        for id_ in sub_process_ids:
            subprocess.call("kill {}".format(id_), shell=True)
            time.sleep(1)

    def _kill_process_windows(self):
        cmd = "taskkill  /F /T /PID {}".format(self.pid)
        ret = subprocess.call(cmd, shell=True)
        return ret



#!/usr/bin/env python
# pylint: disable=attribute-defined-outside-init,
"""
Created on 2020/06/06

@author: yyang
"""
import re
from ctypes import Structure, c_uint8, c_uint64, Union, string_at, addressof, sizeof, Array

from utils import log


class Uint128(Structure):
    _pack_ = 1
    _fields_ = [
        ('LOW', c_uint64),
        ('HIGH', c_uint64),
    ]

    def __init__(self, value=0, **kwargs):
        super(Uint128, self).__init__(**kwargs)
        self.value = value

    def __str__(self):
        return str(self.value)

    def _get_(self):
        return (self.HIGH << 64) | self.LOW

    def _set_(self, value):
        self.HIGH = value >> 64
        self.LOW = value & ((1 << 64) - 1)

    def _del_(self):
        self.LOW = 0
        self.HIGH = 0

    value = property(_get_, _set_, _del_, 'value operation')


class Byte(Structure):
    _pack_ = 1
    _fields_ = [
        ('BYTE', c_uint8),
    ]


class Ctype(object):
    def __init__(self, ctype, name='NAME'):
        self._name = name
        self._ctype = ctype
        self._lines = list()

    @staticmethod
    def _format(indent, name, index=None):
        if index is None:
            return '{}{}'.format(' ' * indent, name)
        return '{}{} [{}]'.format(' ' * indent, name, index)

    @staticmethod
    def _desc(ctype, name):
        try:
            name = getattr(ctype, 'desc').get(name, name)
        except:
            pass
        return name

    def _parser(self, ctype, skip, indent=0, name=''):
        if isinstance(type(ctype), (type(Union), type(Structure))):
            if isinstance(ctype, Uint128):
                self._lines.append([self._format(indent, self._desc(ctype, name)), ctype.value])
                return

            for field in getattr(ctype, '_fields_'):
                name = field[0]
                if [i for i in skip if i in name]:
                    continue
                _ctype = getattr(ctype, name)

                self._parser(_ctype, skip, indent, self._desc(ctype, name))
        elif isinstance(type(ctype), (type(Array), )):
            if isinstance(ctype[0], Byte):
                string = string_at(addressof(ctype), sizeof(ctype)).decode('utf-8', errors='ignore')
                string = re.sub(r'[\x00-\x1f]', '', string).strip()
                if string:
                    self._lines.append([self._format(indent, name), string])
            elif isinstance(type(ctype[0]), (type(Union), type(Structure))):
                for i, value in enumerate(ctype):
                    self._lines.append([self._format(indent, name, i), ' '])
                    self._parser(value, skip, indent + 2)
            else:
                for i, value in enumerate(ctype):
                    self._lines.append([self._format(indent, name, i), value])
        else:
            self._lines.append([self._format(indent, name), ctype])

    def write(self, path, mode='wb'):
        with open(path, mode) as fd:
            fd.write(self._ctype)

    def get(self, name):
        ctype = getattr(self._ctype, name)
        if isinstance(ctype, Uint128):
            return ctype.value
        elif isinstance(type(ctype), type(Array)):
            if isinstance(ctype[0], Byte):
                string = string_at(addressof(ctype), sizeof(ctype)).decode('utf-8', errors='ignore')
                return re.sub(r'[\x00-\x1f]', '', string).strip()
        return ctype

    def dump(self, skip=None, no_rsv=True, only_values=False):
        self._lines = list()
        skip = skip if skip else list()
        if no_rsv:
            skip += ['rsv', 'RSV']
        none_list = (0x0, '', None)

        self._parser(self._ctype, skip)

        name_list = [len(n) for n, v in self._lines if not only_values or v not in none_list] + [len(self._name)]
        max_name_len = max(name_list)
        f_title = '| {{:<{}}} | {{:<34}} | {{:<39}} |'.format(max_name_len)
        f_name = '{{:<{}}}'.format(max_name_len)

        log.DUMP(f_title.format(self._name, 'HEX', 'DEC'))
        log.DUMP('-' * (83 + max_name_len))
        for name, value in self._lines:
            if only_values and value in none_list:
                continue

            line = list()
            line.append(f_name.format(name))
            if isinstance(value, (bool, int, float)):
                line.append("{:<#34x}".format(value))
                line.append("{:<39}".format(value))
            else:
                line.append("{:76}".format(value))

            log.DUMP("| {} |".format(" | ".join(line)))
        log.DUMP('-' * (83 + max_name_len))



# coding=utf-8
from flask import Flask
from flask_restful import Api
from .resource.models.ftp_server import thread_start_ftp_server
from .resource.dut_resource import DutResource

APP = Flask(__name__)
API = Api(APP)


API.add_resource(DutResource, "/dut/status")


if __name__ == '__main__':
    thread_start_ftp_server()
    APP.run(host="0.0.0.0", debug=False)



# coding=utf-8
import sys
import threading
from flask_restful import Resource
from flask_restful import reqparse
from flask_restful import marshal_with
from rest_server.resource.models.helper import resource_fields
from test_framework.dut.dut import Dut
from test_framework.state import State
from utils import log


PARSER = reqparse.RequestParser()
PARSER.add_argument('name')


class DutResource(Resource):

    def __init__(self):
        pass

    @marshal_with(resource_fields, envelope='resource')
    def post(self):
        log.INFO("REST refresh DUT")
        args = PARSER.parse_args()
        self.start_thread_refresh_dut()
        result = {"msg": "start refresh dut", "state": State.PASS}
        return result

    def start_thread_refresh_dut(self):
        thread_p = threading.Thread(target=self.refresh_dut)
        thread_p.setDaemon(True)
        thread_p.start()

    @staticmethod
    def refresh_dut():
        log.INFO("Start refresh DUT")
        dut = Dut()
        dut.refresh()
        log.INFO("End refresh DUT")



#!/usr/bin/env python
# pylint: disable=too-many-locals,too-many-arguments,superfluous-parens,invalid-name,keyword-arg-before-vararg
"""
Created on 2019/4/26

@author: yyang
"""
import re
import fcntl
import math
import os
from ctypes import memmove, sizeof, addressof
from tool.device.nvme.struct import *
from utils import log


def _iowr(_tp, _nr, _sz):
    return (3 << 30) + (sizeof(_sz) << 16) + (ord(_tp) << 8) + _nr


def _iow(_tp, _nr, _sz):
    return (1 << 30) + (sizeof(_sz) << 16) + (ord(_tp) << 8) + _nr


def _io(_tp, _nr):
    return (ord(_tp) << 8) + _nr


class DEVICE:
    _instance = {}
    def __new__(cls, block='/dev/nvme0n1'):
        dev = '{}_{}'.format(block, str(cls))
        #if cls._instance is None:
        if dev not in cls._instance.keys():
            cls._instance[dev] = super().__new__(cls)
        return cls._instance[dev]

    def __init__(self, block='/dev/nvme0n1'):
        self._block = re.findall(r'(nvme\d+n\d+)', block)[0]
        self._cntid = 0
        self._nsid = 0

    @property
    def nsid(self):
        return self._get_nsid()

    @property
    def cntid(self):
        return self._get_cntid()

    def _get_cntid(self):
        path = '/sys/class/nvme'

        for c in os.listdir(path):
            if self._block in os.listdir(os.path.join(path, c)):
                _cntlid = os.path.join(path, c, 'cntid')
                if os.path.exists(_cntlid):
                    return int(open(_cntlid).read().strip())
                return 0

        raise AssertionError('Cannot find cntid of device: {}'.format(self._block))

    def _get_nsid(self):
        path = '/sys/class/nvme'

        for c in os.listdir(path):
            for b in os.listdir(os.path.join(path, c)):
                if b == self._block:
                    return int(open(os.path.join(path, c, b, 'nsid')).read().strip())

        raise AssertionError('Cannot find nsid of device: {}'.format(self._block))


class IOCTL(object):
    NVME_IOCTL_ADMIN_CMD = _iowr('N', 0x41, NVMePassthruCmd)
    NVME_IOCTL_SUBMIT_IO = _iow('N', 0x42, NVMeUserIO)
    NVME_IOCTL_IO_CMD = _iowr('N', 0x43, NVMePassthruCmd)
    NVME_IOCTL_RESET = _io('N', 0x44)
    NVME_IOCTL_SUBSYS_RESET = _io('N', 0x45)
    _instance = {}
    def __new__(cls, cntid=0, nsid=1):
        dev = 'nvme{}n{}_{}'.format(cntid, nsid, str(cls))
        #if cls._instance is None:
        if dev not in cls._instance.keys():
            cls._instance[dev] = super().__new__(cls)
        return cls._instance[dev]

    def __init__(self, cntid=0, nsid=1):
        self._c_fd = -1
        self._b_fd = -1
        self._nsid = nsid
        self._cntid = cntid

        c, b = self._get_dev(cntid=cntid, nsid=nsid)
        self._c_dev = "/dev/{}".format(c)
        self._b_dev = "/dev/{}".format(b)
        log.INFO('%s, %s, cntid: %d, nsid: %d', self._c_dev, self._b_dev, cntid, nsid)

    def __del__(self):
        self._close_c_fd()
        self._close_b_fd()

    @staticmethod
    def _get_b_dev(path, nsid=1):
        for b in os.listdir(path):
            if re.findall(r'nvme\d+[c\d]*n\d+', b):
                _nsid = os.path.join(path, b, 'nsid')
                assert os.path.exists(_nsid), 'File: {} is not exists!'.format(_nsid)
                if int(open(_nsid).read().strip()) == nsid:
                    return re.sub(r'c\d+', '', b)
        return None

    @staticmethod
    def _open(dev):
        assert os.path.exists(dev), 'Device({}) is not exists'.format(dev)
        return os.open(dev, os.O_RDWR)

    @property
    def nsid(self):
        return self._nsid

    @property
    def cntid(self):
        return self._cntid

    @property
    def block_dev(self):
        return self._b_dev

    @property
    def char_fd(self):
        if self._c_fd == -1:
            self._c_fd = self._open(self._c_dev)
        return self._c_fd

    @property
    def block_fd(self):
        if self._b_fd == -1:
            self._b_fd = self._open(self._b_dev)
        return self._b_fd

    def _get_dev(self, cntid=0, nsid=0):
        """
        # 1. nsid=0, return (nvme{cntid}, nvme{cntid}n1)
        2. cntid=0, match nsid only
        3. cntid!=0, match both cntid and nsid
        4. No math, raise AssertionError
        """
        path = '/sys/class/nvme'

        return ["nvme{}".format(cntid), "nvme{}n{}".format(cntid, nsid)]
        # for c in os.listdir(path):
        #     _cntlid = os.path.join(path, c, 'cntlid')
        #
        #     # If specified cntid and cntlid file is exists, should match it.
        #     if cntid != 0 and os.path.exists(_cntlid):
        #         if int(open(_cntlid).read().strip()) != cntid:
        #             continue
        #
        #         b = self._get_b_dev(os.path.join(path, c), nsid=nsid)
        #         return c, b
        #
        #     # If cntid is not specified, only math the nsid.
        #     b = self._get_b_dev(os.path.join(path, c), nsid=nsid)
        #     if b:
        #         return c, b
        #
        # raise AssertionError('No match char device!')

    def _close_c_fd(self):
        if self._c_fd != -1:
            os.close(self._c_fd)
            self._c_fd = -1

    def _close_b_fd(self):
        if self._b_fd != -1:
            os.close(self._b_fd)
            self._b_fd = -1

    def close(self):
        self.__del__()

    def _ioctl(self, fd, request, cmd=None, status=0, timeout=None):
        if timeout and cmd:
            cmd.TIMEOUT = timeout

        _status = fcntl.ioctl(fd, request, cmd) if cmd else fcntl.ioctl(fd, request)

        try:
            assert (_status & 0x7ff) == status, "CQE Status Field Check Failed!"
        except AssertionError:
            log.DUMP('Actual Status: {:#x}, {}'.format(_status, STATUS_FIELD.get(_status & 0x7ff, 'Unknown')))
            log.DUMP("Expect Status: {:#x}, {}".format(status, STATUS_FIELD.get(status & 0x7ff, 'Unknown')))
            if cmd:
                cmd_set = OPCODE.ADMIN if request == self.NVME_IOCTL_ADMIN_CMD else OPCODE.IO
                log.DUMP('Command : {}'.format(cmd_set.get(cmd.OPCODE, 'Unknown')))
                sqe = [(f[0], '{:#x}'.format(getattr(cmd.BITS, f[0]))) for f in getattr(cmd.BITS, '_fields_')]
                log.DUMP('SQE : {}'.format(dict(sqe)))
            raise

        return _status

    def nvme_io(self, cmd, **kwargs):
        return self._ioctl(self.block_fd, self.NVME_IOCTL_SUBMIT_IO, cmd, **kwargs)

    def io_passthru(self, cmd, **kwargs):
        return self._ioctl(self.block_fd, self.NVME_IOCTL_IO_CMD, cmd, **kwargs)

    def admin_passthru(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_ADMIN_CMD, cmd, **kwargs)

    def reset(self, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_RESET, **kwargs)

    def subsys_reset(self, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_SUBSYS_RESET, **kwargs)

    @staticmethod
    def set_expected_status_field(status):
        assert False, 'Not support, status={:#x}!'.format(status)


class ADMIN(IOCTL):
    def abort(self, sqid, cid, **kwargs):
        cdw10 = AbortDword10()
        cdw10.SQID = sqid
        cdw10.CID = cid

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.ABORT
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs), cmd.RESULT

    def create_io_cq(self, qid, qsize, pc=1, ien=1, iv=0, **kwargs):
        cdw10 = CreateIOCQDword10()
        cdw10.QID = qid
        cdw10.QSIZE = qsize - 1

        cdw11 = CreateIOCQDword11()
        cdw11.PC = pc
        cdw11.IEN = ien
        cdw11.IV = iv

        dptr = (c_uint32 * (4 * qsize))()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.CREATE_IO_CQ
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)

        return self.admin_passthru(cmd, **kwargs)

    def create_io_sq(self, qid, qsize, pc=1, qprio=1, cqid=0, **kwargs):
        cdw10 = CreateIOCQDword10()
        cdw10.QID = qid
        cdw10.QSIZE = qsize - 1

        cdw11 = CreateIOSQDword11()
        cdw11.PC = pc
        cdw11.QPRIO = qprio
        cdw11.CQID = cqid

        dptr = (c_uint32 * (16 * qsize))()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.CREATE_IO_SQ
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)

        return self.admin_passthru(cmd, **kwargs)

    def delete_io_cq(self, qid, **kwargs):
        cdw10 = DeleteIOCQDword10()
        cdw10.QID = qid

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DELETE_IO_CQ
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def delete_io_sq(self, qid, **kwargs):
        cdw10 = DeleteIOSQDword10()
        cdw10.QID = qid

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DELETE_IO_SQ
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def device_self_test(self, stc, nsid=1, **kwargs):
        cdw10 = DeviceSelfTest()
        cdw10.STC = stc

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DEVICE_SELF_TEST
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def directive_send(self, dptr, doper=0, dtype=0, dspec=0, cdw12=0, nsid=1, **kwargs):
        cdw11 = DirectiveDW11()
        cdw11.DOPER = doper
        cdw11.DTYPE = dtype
        cdw11.DSPEC = dspec

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DIRECTIVE_SEND
        cmd.NSID = nsid
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = (sizeof(dptr) >> 2) - 1
        cmd.CDW[11] = cdw11.DWORD
        cmd.CDW[12] = cdw12

        return self.admin_passthru(cmd, **kwargs)

    def directive_receive(self, dptr, doper=0, dtype=0, dspec=0, cdw12=0, nsid=1, **kwargs):
        cdw11 = DirectiveDW11()
        cdw11.DOPER = doper
        cdw11.DTYPE = dtype
        cdw11.DSPEC = dspec

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DIRECTIVE_RECEIVE
        cmd.NSID = nsid
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = (sizeof(dptr) >> 2) - 1
        cmd.CDW[11] = cdw11.DWORD
        cmd.CDW[12] = cdw12

        return self.admin_passthru(cmd, **kwargs)

    def nvme_mi_receive(self, dptr, nmsp0, **kwargs):
        cdw10 = NVMeMIDW10()
        cdw10.NMSP0 = nmsp0

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.NVME_MI_RECEIVE
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs), cmd.RESULT

    def nvme_mi_send(self, dptr, nmsp0, **kwargs):
        cdw10 = NVMeMIDW10()
        cdw10.NMSP0 = nmsp0

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.NVME_MI_SEND
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def fw_commit(self, ca=3, fs=1, bpid=0, **kwargs):
        cdw10 = FWCommitDW10()
        cdw10.BPID = bpid
        cdw10.CA = ca
        cdw10.FS = fs

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.FIRMWARE_COMMIT
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def fw_download(self, filename, xfer=0, offset=0, align=4096, **kwargs):
        file_size = os.path.getsize(filename)
        length = math.ceil(file_size / 4096) * 4096
        dptr = (c_uint8 * length)()
        memmove(dptr, open(filename, "rb").read(), file_size)
        addr = addressof(dptr)

        offset <<= 2
        if xfer == 0 or xfer % align:
            xfer = 4096

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.FIRMWARE_DOWNLOAD

        while file_size > 0:
            xfer = min(xfer, file_size)

            cmd.CDW[10] = (xfer >> 2) - 1
            cmd.CDW[11] = offset >> 2
            cmd.DATA_LEN = xfer
            cmd.DATA = addr

            status = self.admin_passthru(cmd, **kwargs)
            if status:
                return status

            addr += xfer
            file_size -= xfer
            offset += xfer

        return 0

    def get_features(self, fid, sel, cdw11=0, nsid=1, **kwargs):
        cdw10 = GetFeaturesDword10()
        cdw10.FID = fid
        cdw10.SEL = sel

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.GET_FEATURES
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11

        if fid in GET_FEATURE_ENTRY:
            length = 4096 // sizeof(GET_FEATURE_ENTRY[fid])
            dptr = (GET_FEATURE_ENTRY[fid] * length)()
            cmd.DATA = addressof(dptr)
            cmd.DATA_LEN = 4096
        else:
            dptr = None

        status = self.admin_passthru(cmd, **kwargs)
        cdw0 = GET_FEATURE_ID.get(fid, CDW)()
        cdw0.DWORD = cmd.RESULT
        return status, cdw0, dptr

    def get_log(self, dptr, lid=0x1, lpol=0, lpou=0, lsp=0, xfer=4096, nsid=0xffffffff, **kwargs):
        addr = addressof(dptr)
        size = sizeof(dptr)

        cdw10 = GetLogPagDword10()
        cdw11 = GetLogPagDword11()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.GET_LOG_PAGE
        cmd.NSID = nsid

        while size > 0:
            xfer = min(xfer, size)
            numd = (xfer >> 2) - 1

            cdw10.LID = lid
            cdw10.LSP = lsp
            cdw10.NUMDL = numd & 0xffff
            cdw11.NUMDU = numd >> 16

            cmd.CDW[10] = cdw10.DWORD
            cmd.CDW[11] = cdw11.DWORD
            cmd.CDW[12] = lpol
            cmd.CDW[13] = lpou
            cmd.DATA = addr
            cmd.DATA_LEN = xfer

            status = self.admin_passthru(cmd, **kwargs)
            if status:
                return status

            addr += xfer
            size -= xfer
            lpou, lpol = divmod(lpol + xfer, 2 ** 32)

        return 0

    def identify(self, nsid=0, cns=0, cntid=0, **kwargs):
        cdw10 = IdentifyDword10()
        cdw10.CNS = cns
        cdw10.CNTID = cntid

        dptr = IDENTIFY_CNS.get(cns, (c_uint8 * 4096))()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.IDENTIFY
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)

        return self.admin_passthru(cmd, **kwargs), dptr

    def namespace_attachment(self, sel, cntid_list, nsid=1, **kwargs):
        cdw10 = AttachNSDword10()
        cdw10.SEL = sel

        dptr = ControllerList()
        dptr.NUM = len(cntid_list)
        for i, cntid in enumerate(cntid_list):
            dptr.ID[i] = cntid

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.NAMESPACE_ATTACH
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        return self.admin_passthru(cmd, **kwargs)

    def namespace_management(self, sel, nsze=0, ncap=0, flbas=0, dps=0, nmic=0, nsid=0, **kwargs):
        cdw10 = NamespaceManagementDword10()
        cdw10.SEL = sel

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.NAMESPACE_MGT
        cmd.CDW[10] = cdw10.DWORD

        if sel == 0:  # Create
            dptr = NamespaceManagementDataStructure()
            dptr.NSZE = nsze
            dptr.NCAP = ncap
            dptr.FLBAS = flbas
            dptr.DPS = dps
            dptr.NMIC = nmic

            cmd.NSID = 0
            cmd.DATA = addressof(dptr)
            cmd.DATA_LEN = sizeof(dptr)
        elif sel == 1:  # Delete
            cmd.NSID = nsid

        return self.admin_passthru(cmd, **kwargs), cmd.RESULT

    def set_features(self, fid, save=0, cdw11=0, cdw12=0, cdw13=0, cdw14=0, cdw15=0, dptr=None, nsid=0xffffffff,
                     **kwargs):
        cdw10 = SetFeaturesDword10()
        cdw10.FID = fid
        cdw10.SV = save

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.SET_FEATURES
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11
        cmd.CDW[12] = cdw12
        cmd.CDW[13] = cdw13
        cmd.CDW[14] = cdw14
        cmd.CDW[15] = cdw15
        if dptr:
            cmd.DATA = addressof(dptr)
            cmd.DATA_LEN = sizeof(dptr)

        return self.admin_passthru(cmd, **kwargs)

    def sanitize(self, sanact=0, ause=0, owpass=0, oipbp=0, ndas=0, ovpra=0, nsid=0, **kwargs):
        cmd = NVMePassthruCmd()
        cdw10 = SanitizeDW10()
        cmd.OPCODE = OPCODE.SANITIZE
        cmd.NSID = nsid
        cdw10.SANACT = sanact
        cdw10.AUSE = ause
        cdw10.OWPASS = owpass
        cdw10.OIPBP = oipbp
        cdw10.NDAS = ndas

        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = ovpra

        return self.admin_passthru(cmd, **kwargs)

    def virtualization_management(self, cntlid, rt, act, nr=0, **kwargs):
        cmd = NVMePassthruCmd()
        cdw10 = VirtualizationManagementDword10()
        cdw11 = VirtualizationManagementDword11()

        cmd.OPCODE = OPCODE.VIRT_MANAGEMENT
        cdw10.ACT = act
        cdw10.RT = rt
        cdw10.CNTLID = cntlid
        cdw11.NR = nr

        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def format(self, nsid=1, lbaf=0, mset=0, pif=0, pil=0, ses=0, **kwargs):
        """
        Admin Command Set - Format NVM command
        :param lbaf:    LBA Format (LBAF)
        :param nsid:    NameSpace ID (NSID)
        :param mset:    Metadata Settings (MSET)
        :param pif:     Protection Information (PIF)
        :param pil:     Protection Information Location (PIL)
        :param ses:     Secure Erase Settings (SES)
        :return:        Status Field Of CQE
        """
        cdw10 = FormatDword10()
        cdw10.LBAF = lbaf
        cdw10.MSET = mset
        cdw10.PIF = pif
        cdw10.PIL = pil
        cdw10.SES = ses

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.FORMAT_NVM
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD

        return self.admin_passthru(cmd, **kwargs)

    def security_receive(self, nsid=1, nssf=0, spsp0=0, spsp1=0, secp=0, atl=0, **kwargs):
        cdw10 = SecurityReceiveDword10()
        cdw10.NSSF = nssf
        cdw10.SPSP0 = spsp0
        cdw10.SPSP1 = spsp1
        cdw10.SECP = secp

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.SECURITY_RECV
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = atl

        return self.admin_passthru(cmd, **kwargs)

    def security_send(self, nsid=1, nssf=0, spsp0=0, spsp1=0, secp=0, atl=0, **kwargs):
        cdw10 = SecuritySendDword10()
        cdw10.NSSF = nssf
        cdw10.SPSP0 = spsp0
        cdw10.SPSP1 = spsp1
        cdw10.SECP = secp

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.SECURITY_SEND
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = atl

        return self.admin_passthru(cmd, **kwargs)


class IO(IOCTL):
    def compare(self, slba, nlb, dptr, mptr=None, nsid=1, prinfo=0, fua=0, lr=0, reftag=0, apptag=0, appmask=0,
                **kwargs):
        cdw12 = RWDword12()
        cdw12.NLB = nlb - 1
        cdw12.PRINFO = prinfo
        cdw12.FUA = fua
        cdw12.LR = lr

        cmd = NVMeUserIO()
        cmd.OPCODE = OPCODE.COMPARE
        cmd.NSID = nsid
        cmd.SLBA = slba
        cmd.NBLOCKS = cdw12.NBLOCKS
        cmd.CONTROL = cdw12.CONTROL
        cmd.DATA = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.REFTAG = reftag
        cmd.APPTAG = apptag
        cmd.APPMASK = appmask

        return self.nvme_io(cmd, **kwargs)

    def dataset_management(self, nr, dptr, idr=0, idw=0, ad=1, nsid=1, **kwargs):
        cdw10 = DsmDword10()
        cdw10.NR = nr - 1

        cdw11 = DsmDword11()
        cdw11.IDR = idr
        cdw11.IDW = idw
        cdw11.AD = ad

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DATASET_MGT
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)

        return self.io_passthru(cmd, **kwargs)

    def flush(self, nsid=1, **kwargs):
        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.FLUSH
        cmd.NSID = nsid

        return self.io_passthru(cmd, **kwargs)

    def read(self, slba, nlb, dptr, mptr=None, prinfo=0, fua=0, lr=0, dsm=0, reftag=0, apptag=0, appmask=0, **kwargs):
        cdw12 = RWDword12()
        cdw12.NLB = nlb - 1
        cdw12.PRINFO = prinfo
        cdw12.FUA = fua
        cdw12.LR = lr

        cmd = NVMeUserIO()
        cmd.OPCODE = OPCODE.READ
        cmd.SLBA = slba
        cmd.NBLOCKS = cdw12.NBLOCKS
        cmd.CONTROL = cdw12.CONTROL
        cmd.DATA = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.DSMGMT = dsm
        cmd.REFTAG = reftag
        cmd.APPTAG = apptag
        cmd.APPMASK = appmask

        return self.nvme_io(cmd, **kwargs)

    def reservation_acquire(self, racqa, iekey, rtype, crkey, prkey, nsid=1, **kwargs):
        cdw10 = ReservationAcquire()
        cdw10.RACQA = racqa
        cdw10.IEKEY = iekey
        cdw10.RTYPE = rtype

        dptr = ReservationAcquireDataStructure()
        dptr.CRKEY = crkey
        dptr.PRKEY = prkey

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.RESERVATION_ACQUIRE
        cmd.NSID = nsid
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = cdw10.DWORD

        return self.io_passthru(cmd, **kwargs)

    def reservation_register(self, rrega, iekey, cptpl, crkey, nrkey, nsid=1, **kwargs):
        cdw10 = ReservationRegister()
        cdw10.RREGA = rrega
        cdw10.IEKEY = iekey
        cdw10.CPTPL = cptpl

        dptr = ReservationRegisterDataStructure()
        dptr.CRKEY = crkey
        dptr.NRKEY = nrkey

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.RESERVATION_REGISTER
        cmd.NSID = nsid
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = cdw10.DWORD

        return self.io_passthru(cmd, **kwargs)

    def reservation_release(self, rrela, iekey, rtype, crkey, nsid=1, **kwargs):
        cdw10 = ReservationRelease()
        cdw10.RRELA = rrela
        cdw10.IEKEY = iekey
        cdw10.RTYPE = rtype

        dptr = ReservationReleaseDataStructure()
        dptr.CRKEY = crkey

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.RESERVATION_RELEASE
        cmd.NSID = nsid
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = cdw10.DWORD

        return self.io_passthru(cmd, **kwargs)

    def reservation_report(self, eds, nsid=1, **kwargs):
        if eds == 0:
            dptr = ReservationStatusDataStructure()
        else:
            dptr = ReservationStatusExtendedDataStructure()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.RESERVATION_REPORT
        cmd.NSID = nsid
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        cmd.CDW[10] = (sizeof(dptr) >> 2) - 1
        cmd.CDW[11] = eds & 0x1

        return self.io_passthru(cmd, **kwargs), dptr

    def write(self, slba, nlb, dptr, mptr=None, dtype=0, prinfo=0, fua=0, lr=0, dsm=0, dspec=0, reftag=0, apptag=0,
              appmask=0, **kwargs):
        cdw12 = RWDword12()
        cdw12.NLB = nlb - 1
        cdw12.DTYPE = dtype
        cdw12.PRINFO = prinfo
        cdw12.FUA = fua
        cdw12.LR = lr

        cdw13 = RWDword13()
        cdw13.DSM = dsm
        cdw13.DSPEC = dspec

        cmd = NVMeUserIO()
        cmd.OPCODE = OPCODE.WRITE
        cmd.SLBA = slba
        cmd.NBLOCKS = cdw12.NBLOCKS
        cmd.CONTROL = cdw12.CONTROL
        cmd.DATA = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.DSMGMT = cdw13.DWORD
        cmd.REFTAG = reftag
        cmd.APPTAG = apptag
        cmd.APPMASK = appmask

        return self.nvme_io(cmd, **kwargs)

    def write_uncorrectable(self, slba, nlb, nsid=1, **kwargs):
        cdw12 = RWDword12()
        cdw12.NLB = nlb - 1

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.WRITE_UNCORRECTABLE
        cmd.NSID = nsid
        cmd.CDW[10] = slba & 0xffffffff
        cmd.CDW[11] = slba >> 32
        cmd.CDW[12] = cdw12.DWORD

        return self.io_passthru(cmd, **kwargs)

    def write_zeroes(self, slba, nlb, nsid=1, deac=0, prinfo=0, fua=0, lr=0, reftag=0, apptag=0, appmask=0, **kwargs):
        cdw12 = RWDword12()
        cdw12.NLB = nlb - 1
        cdw12.DEAC = deac
        cdw12.PRINFO = prinfo
        cdw12.FUA = fua
        cdw12.LR = lr

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.WRITE_ZEROES
        cmd.NSID = nsid
        cmd.CDW[10] = slba & 0xffffffff
        cmd.CDW[11] = slba >> 32
        cmd.CDW[12] = cdw12.DWORD
        cmd.CDW[14] = reftag
        cmd.CDW[15] = apptag | (appmask << 16)

        return self.io_passthru(cmd, **kwargs)


class PIOCTL(IOCTL):
    NVME_IOCTL_ADMIN_CMD = _iowr('E', 0x01, NVMePassthruCmd)
    NVME_IOCTL_IO = _iowr('E', 0x02, PNVMeIOCmd)
    NVME_IOCTL_AIO = _iowr('E', 0x03, PNVMeIOCmd)
    NVME_IOCTL_ATOMIC = _iowr('E', 0x04, PNVMeIOCmd)
    NVME_IOCTL_PI = _iowr('E', 0x05, PNVMeIOCmd)
    NVME_IOCTL_SGL = _iowr('E', 0x06, PNVMeIOCmd)
    NVME_IOCTL_CRT_CQ = _iowr('E', 0x07, NVMePassthruCmd)
    NVME_IOCTL_DEL_CQ = _iowr('E', 0x08, NVMePassthruCmd)
    NVME_IOCTL_CRT_SQ = _iowr('E', 0x09, NVMePassthruCmd)
    NVME_IOCTL_DEL_SQ = _iowr('E', 0x0a, NVMePassthruCmd)

    NVME_IOCTL_SQ = _iowr('G', 0x01, c_uint32)
    NVME_IOCTL_CQ = _iowr('G', 0x02, c_uint32)
    NVME_IOCTL_DEBUG = _iowr('G', 0x03, c_uint32)

    def __init__(self, device='/dev/pnvme', cntid=0, nsid=1):
        super().__init__(cntid, nsid)
        self._c_fd = -1
        self._c_dev = device

    def __del__(self):
        self._close_c_fd()

    def close(self):
        self.__del__()

    def io_passthru(self, cmd, **kwargs):
        assert False, 'Not support io_passthru for PIOCTL'

    def reset(self, **kwargs):
        assert False, 'Not support reset for PIOCTL'

    def subsys_reset(self, **kwargs):
        assert False, 'Not support subsys_reset for PIOCTL'

    def nvme_debug(self, cmd):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_DEBUG, cmd)

    def nvme_get_sq(self, cmd):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_SQ, cmd)

    def nvme_get_cq(self, cmd):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_CQ, cmd)

    def nvme_io(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_IO, cmd, **kwargs)

    def nvme_pi(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_PI, cmd, **kwargs)

    def nvme_sgl(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_SGL, cmd, **kwargs)

    def nvme_aio(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_AIO, cmd, **kwargs)

    def nvme_atomic(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_ATOMIC, cmd, **kwargs)

    def admin_passthru(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_ADMIN_CMD, cmd, **kwargs)

    def nvme_crt_cq(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_CRT_CQ, cmd, **kwargs)

    def nvme_del_cq(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_DEL_CQ, cmd, **kwargs)

    def nvme_crt_sq(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_CRT_SQ, cmd, **kwargs)

    def nvme_del_sq(self, cmd, **kwargs):
        return self._ioctl(self.char_fd, self.NVME_IOCTL_DEL_SQ, cmd, **kwargs)

    def identify(self, nsid=1, cns=1, cntid=0, **kwargs):
        cdw10 = IdentifyDword10()
        cdw10.CNS = cns
        cdw10.CNTID = cntid
        dptr = ControllerDataStructure()
        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.IDENTIFY
        cmd.NSID = nsid
        cmd.CDW[10] = cdw10.DWORD
        cmd.ADDR = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        return self.admin_passthru(cmd, **kwargs), dptr

    def create_io_cq(self, qid, qsize, pc=1, ien=1, iv=0, **kwargs):
        cdw10 = CreateIOCQDword10()
        cdw10.QID = qid
        cdw10.QSIZE = qsize - 1

        cdw11 = CreateIOCQDword11()
        cdw11.PC = pc
        cdw11.IEN = ien
        cdw11.IV = iv

        dptr = (c_uint32 * (4 * qsize))()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.CREATE_IO_CQ
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)

        return self.nvme_crt_cq(cmd, **kwargs)

    def create_io_sq(self, qid, qsize, pc=1, qprio=1, cqid=0, **kwargs):
        cdw10 = CreateIOCQDword10()
        cdw10.QID = qid
        cdw10.QSIZE = qsize - 1

        cdw11 = CreateIOSQDword11()
        cdw11.PC = pc
        cdw11.QPRIO = qprio
        cdw11.CQID = cqid

        dptr = (c_uint32 * (16 * qsize))()

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.CREATE_IO_SQ
        cmd.CDW[10] = cdw10.DWORD
        cmd.CDW[11] = cdw11.DWORD
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)

        return self.nvme_crt_sq(cmd, **kwargs)

    def delete_io_cq(self, qid, **kwargs):
        cdw10 = DeleteIOCQDword10()
        cdw10.QID = qid

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DELETE_IO_CQ
        cmd.CDW[10] = cdw10.DWORD

        return self.nvme_del_cq(cmd, **kwargs)

    def delete_io_sq(self, qid, **kwargs):
        cdw10 = DeleteIOSQDword10()
        cdw10.QID = qid

        cmd = NVMePassthruCmd()
        cmd.OPCODE = OPCODE.DELETE_IO_SQ
        cmd.CDW[10] = cdw10.DWORD

        return self.nvme_del_sq(cmd, **kwargs)

    def write(self, slba, nlb, dptr, mptr=None, qid=1, nsid=1, dsm=0, **kwargs):
        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.WRITE
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.PRP1 = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.DSMGMT = dsm
        return self.nvme_io(cmd, **kwargs)

    def read(self, slba, nlb, dptr, mptr=None, qid=1, nsid=1, dsm=0, **kwargs):

        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.READ
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.PRP1 = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.DSMGMT = dsm
        return self.nvme_io(cmd, **kwargs)

    def write_sgl(self, slba, nlb, d_num, m_num, dptr, dsglptr, mptr=None, msglptr=None, qid=1, nsid=1, ctrl=0,
                  flags=0x40, doff=0, moff=0, **kwargs):
        cmd = NVMeSGLIO()
        cmd.OPCODE = OPCODE.WRITE
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.PRP1 = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.CTRL = ctrl
        cmd.FLAGS = flags
        cmd.DATA_OFF = doff
        cmd.META_OFF = moff
        cmd.DATA_DES_NUM = d_num if d_num >= 1 else 1
        cmd.META_DES_NUM = m_num
        cmd.SGL_DATA_PTR = addressof(dsglptr)
        cmd.SGL_META_PTR = addressof(msglptr) if mptr else 0
        return self.nvme_sgl(cmd, **kwargs)

    def read_sgl(self, slba, nlb, d_num, m_num, dptr, dsglptr, mptr=None, msglptr=None, qid=1, nsid=1, ctrl=0,
                 flags=0x40, doff=0, moff=0, **kwargs):
        cmd = NVMeSGLIO()
        cmd.OPCODE = OPCODE.READ
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.PRP1 = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.CTRL = ctrl
        cmd.FLAGS = flags
        cmd.DATA_OFF = doff
        cmd.META_OFF = moff
        cmd.DATA_DES_NUM = d_num if d_num >= 1 else 1
        cmd.META_DES_NUM = m_num
        cmd.SGL_DATA_PTR = addressof(dsglptr)
        cmd.SGL_META_PTR = addressof(msglptr) if mptr else 0
        return self.nvme_sgl(cmd, **kwargs)

    def write_async(self, slba, nlb, qid=1, nsid=1, **kwargs):
        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.WRITE
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        return self.nvme_aio(cmd, **kwargs)

    def read_async(self, slba, nlb, qid=1, nsid=1, **kwargs):
        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.READ
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        return self.nvme_aio(cmd, **kwargs)

    def write_atomic(self, slba, nlb, qid=1, nsid=1, pat=0, **kwargs):
        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.WRITE
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.CTRL = pat
        return self.nvme_atomic(cmd, **kwargs)

    def read_atomic(self, slba, nlb, qid=1, nsid=1, **kwargs):
        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.READ
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        return self.nvme_atomic(cmd, **kwargs)

    def async_trace(self):
        dptr = AsyncTrace()
        cmd = PNVMeDebug()
        cmd.OPCODE = DebugOPCODE.ASYNC_TRACE
        cmd.DATA = addressof(dptr)
        ret = self.nvme_debug(cmd)
        return ret, dptr

    def get_cq(self, qid=0):
        return self.nvme_get_cq(qid)

    def get_sq(self, qid=0):
        return self.nvme_get_sq(qid)

    def write_pi(self, slba, nlb, dptr, mptr=None, qid=1, nsid=1, ctrl=0, reftag=0, apptag=0, appmask=0, **kwargs):
        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.WRITE
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.CTRL = ctrl
        cmd.PRP1 = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.REFTAG = reftag
        cmd.APPTAG = apptag
        cmd.APPMASK = appmask
        return self.nvme_pi(cmd, **kwargs)

    def read_pi(self, slba, nlb, dptr, mptr=None, qid=1, nsid=1, ctrl=0, reftag=0, apptag=0, appmask=0, **kwargs):

        cmd = PNVMeIOCmd()
        cmd.OPCODE = OPCODE.READ
        cmd.SLBA = slba
        cmd.NLB = nlb - 1
        cmd.QID = qid
        cmd.NSID = nsid
        cmd.CTRL = ctrl
        cmd.PRP1 = addressof(dptr)
        cmd.META = addressof(mptr) if mptr else 0
        cmd.REFTAG = reftag
        cmd.APPTAG = apptag
        cmd.APPMASK = appmask
        return self.nvme_pi(cmd, **kwargs)



#!/usr/bin/env python
# pylint: disable=superfluous-parens,missing-docstring,too-few-public-methods
"""
Created on Apr 19, 2019

@author: yyang
"""
from ctypes import Structure, c_uint8, c_uint16, c_uint32, c_uint64, Union
from tool.device.nvme.ctype import Uint128, Byte


class ErrorInformation(Structure):
    """Get Log Page - Error Information (Log Identifier 01h)"""
    _pack_ = 1
    _fields_ = [
        ('COUNT', c_uint64),  # Error Count
        ('SQID', c_uint16),  # Submission Queue ID
        ('CID', c_uint16),  # Command ID
        ('STATUS', c_uint16),  # Status Field
        ('PEL', c_uint16),  # Parameter Error Location
        ('LBA', c_uint64),  # LBA
        ('NSID', c_uint32),  # Namespace
        ('VSIA', c_uint8),  # Vendor Specific Information Available
        ('RSV0', c_uint8 * 3),
        ('CSI', c_uint64),  # Command Specific Information
        ('RSV1', c_uint8 * 24),
    ]
    desc = {
        'COUNT': 'Error Count', 'SQID': 'Submission Queue ID', 'CID': 'Command ID', 'STATUS': 'Status Field',
        'PEL': 'Parameter Error Location', 'LBA': 'LBA', 'NSID': 'Namespace',
        'VSIA': 'Vendor Specific Information Available', 'CSI': 'Command Specific Information'
    }


class SmartHealthInformation(Structure):
    """Get Log Page - SMART/Health Information (Log Identifier 02h)"""
    _pack_ = 1
    _fields_ = [
        ('CW', c_uint8),  # Critical Warning
        ('CT', c_uint16),  # Composite Temperature
        ('AS', c_uint8),  # Available Spare
        ('AST', c_uint8),  # Available Spare Threshold
        ('PU', c_uint8),  # Percentage Used
        ('EGCWS', c_uint8),  # Endurance Group Critical Warning Summary
        ('RSV0', c_uint8 * 25),
        ('DUR', Uint128),  # Data Units Read
        ('DUW', Uint128),  # Data Units Write
        ('HRC', Uint128),  # Host Read Commands
        ('HWC', Uint128),  # Host Write Commands
        ('CBT', Uint128),  # Controller Busy Time
        ('PC', Uint128),  # Power Cycles
        ('POH', Uint128),  # Power On Hours
        ('US', Uint128),  # Unsafe Shutdowns
        ('MDIE', Uint128),  # Media and Data Integrity Errors
        ('NEILE', Uint128),  # Number of Error Information Log Entries
        ('WCTT', c_uint32),  # Warning Composite Temperature Time
        ('CCTT', c_uint32),  # Critical Composite Temperature Time
        ('TS', c_uint16 * 8),  # Temperature Sensor 1 ~ 8
        ('TMT1TC', c_uint32),  # Thermal Management Temperature 1 Transition Count
        ('TMT2TC', c_uint32),  # Thermal Management Temperature 2 Transition Count
        ('TTFTMT1', c_uint32),  # Total Time For Thermal Management Temperature 1
        ('TTFTMT2', c_uint32),  # Total Time For Thermal Management Temperature 2
        ('RSV1', c_uint8 * 280),
    ]
    desc = {
        'CW': 'Critical Warning', 'CT': 'Composite Temperature', 'AS': 'Available Spare',
        'AST': 'Available Spare Threshold', 'PU': 'Percentage Used',
        'EGCWS': 'Endurance Group Critical Warning Summary', 'DUR': 'Data Units Read',
        'DUW': 'Data Units Write', 'HRC': 'Host Read Commands', 'HWC': 'Host Write Commands',
        'CBT': 'Controller Busy Time', 'PC': 'Power Cycles', 'POH': 'Power On Hours',
        'US': 'Unsafe Shutdowns', 'MDIE': 'Media and Data Integrity Errors',
        'NEILE': 'Number of Error Information Log Entries', 'WCTT': 'Warning Composite Temperature Time',
        'CCTT': 'Critical Composite Temperature Time', 'TS': 'Temperature Sensor',
        'TMT1TC': 'Thermal Management Temperature 1 Transition Count',
        'TMT2TC': 'Thermal Management Temperature 2 Transition Count',
        'TTFTMT1': 'Total Time For Thermal Management Temperature 1',
        'TTFTMT2': 'Total Time For Thermal Management Temperature 2',
    }


class FirmwareSlotInformation(Structure):
    """Get Log Page - Firmware Slot Information (Log Identifier 03h)"""
    _pack_ = 1
    _fields_ = [
        ('AFI', c_uint8),  # Active Firmware Info (AFI)
        ('RSV0', c_uint8 * 7),
        ('FRS1', Byte * 8),  # Firmware Revision for Slot 1 (FRS1)
        ('FRS2', Byte * 8),  # Firmware Revision for Slot 2 (FRS2)
        ('FRS3', Byte * 8),  # Firmware Revision for Slot 3 (FRS3)
        ('FRS4', Byte * 8),  # Firmware Revision for Slot 4 (FRS4)
        ('FRS5', Byte * 8),  # Firmware Revision for Slot 5 (FRS5)
        ('FRS6', Byte * 8),  # Firmware Revision for Slot 6 (FRS6)
        ('FRS7', Byte * 8),  # Firmware Revision for Slot 7 (FRS7)
        ('RSV1', c_uint8 * 4032)
    ]
    desc = {
        'AFI': 'Active Firmware Info', 'FRS1': 'Firmware Revision for Slot 1',
        'FRS2': 'Firmware Revision for Slot 2', 'FRS3': 'Firmware Revision for Slot 3',
        'FRS4': 'Firmware Revision for Slot 4', 'FRS5': 'Firmware Revision for Slot 5',
        'FRS6': 'Firmware Revision for Slot 6', 'FRS7': 'Firmware Revision for Slot 7',
    }


class ChangedNamespaceList(Structure):
    """Get Log Page - Changed Namespace list data structure"""
    _pack_ = 1
    _fields_ = [
        ('ID', c_uint64 * 1024)
    ]
    desc = {
        'ID': 'Identifier Index',
    }


class CommandEffectDataStructure(Structure):
    """Get Log Page - Command Effect Data Structure (Log Identifier 05h)"""
    _pack_ = 1
    _fields_ = [
        ('csupp', c_uint32, 1),
        ('lbcc', c_uint32, 1),
        ('ncc', c_uint32, 1),
        ('nic', c_uint32, 1),
        ('ccc', c_uint32, 1),
        ('rsv0', c_uint32, 11),
        ('cse', c_uint32, 3),
        ('rsv1', c_uint32, 13),
    ]


class CommandsSupportAndEffects(Structure):
    _pack_ = 1
    _fields_ = [
        ('ACS', CommandEffectDataStructure * 256),
        ('IOCS', CommandEffectDataStructure * 256),
        ('RSV', c_uint8 * 2048),
    ]
    desc = {
        'ACS': 'Admin Command Supported', 'IOCS': 'I/O Command Supported',
    }


class SelfTestResult(Structure):
    """Get Log Page - Device Self-test data structure"""
    _pack_ = 1
    _fields_ = [
        ('DSTS', c_uint8),
        ('SN', c_uint8),
        ('VDI', c_uint8),
        ('RSV', c_uint8),
        ('POH', c_uint64),
        ('NSID', c_uint32),
        ('FLBA', c_uint64),
        ('SCT', c_uint8),
        ('SC', c_uint8),
        ('VS', c_uint16)
    ]
    desc = {
        'DSTS': 'device selt-test status', 'SN': 'segment number', 'VDI': 'valid diagnostic information',
        'POH': 'power on hours', 'NSID': 'namespace identifier', 'FLBA': 'failing lba',
        'SCT': 'status code type', 'SC': 'status code', 'VS': 'vendor specific',
    }


class SelfTestLog(Structure):
    """Get Log Page - Device Self-test (Log Identifier 06h)"""
    _pack_ = 1
    _fields_ = [
        ('CDSTO', c_uint8),  # current device self-test operation
        ('CDSTC', c_uint8),  # current device self-test completion
        ('RSV', c_uint16),
        ('RESULT', SelfTestResult * 20)  # self test result data structure
    ]
    desc = {
        'CDSTO': 'current device self-test operation', 'CDSTC': 'current device self-test completion',
        'RESULT': 'self test result data structure',
    }


class TelemetryDataBlock(Structure):
    """Get Log Page - Telemetry Data Block Structure"""
    _pack_ = 1
    _fields_ = [
        ('DATA', c_uint8 * 512),  # Telemetry Data Block
    ]
    desc = {
        'DATA': 'Telemetry Data Block',
    }


class TelemetryHostInitiated(Structure):
    """Get Log Page - Telemetry Host-Initiated Log(Log Identifier 07h)"""
    _pack_ = 1
    _fields_ = [
        ('LID', c_uint8),  # Log Identifier, This field shall be set to 07h.
        ('RSV0', c_uint32),  # Reserved 0
        ('IEEE', c_uint8 * 3),  # IEEE OUI Identifier
        ('THIDA1LB', c_uint16),  # Telemetry Host-Initiated Data Area 1 Last Block
        ('THIDA2LB', c_uint16),  # Telemetry Host-Initiated Data Area 2 Last Block
        ('THIDA3LB', c_uint16),  # Telemetry Host-Initiated Data Area 3 Last Block
        ('RSV1', c_uint8 * 368),  # Reserved 1
        ('TCIDA', c_uint8),  # Telemetry Controller-Initiated Data Available
        ('TCIDGN', c_uint8),  # Telemetry Controller-Initiated Data Generation Number
        ('RID', c_uint8 * 128),  # Reason Identifier
        ('THIDB', TelemetryDataBlock * 7),  # Telemetry Host-Initiated Data Block 1 - n
    ]
    desc = {
        'LID': 'Log Identifier', 'IEEE': 'IEEE OUI Identifier',
        'THIDA1LB': 'Telemetry Host-Initiated Data Area 1 Last Block',
        'THIDA2LB': 'Telemetry Host-Initiated Data Area 2 Last Block',
        'THIDA3LB': 'Telemetry Host-Initiated Data Area 3 Last Block',
        'TCIDA': 'Telemetry Controller-Initiated Data Available',
        'TCIDGN': 'Telemetry Controller-Initiated Data Generation Number',
        'RID': 'Reason Identifier', 'THIDB': 'Telemetry Host-Initiated Data Block',
    }


class TelemetryControllerInitiated(Structure):
    """Get Log Page - Telemetry Controller-Initiated Log(Log Identifier 08h)"""
    _pack_ = 1
    _fields_ = [
        ('LID', c_uint8),  # Log Identifier, This field shall be set to 08h.
        ('RSV0', c_uint32),  # Reserved 0
        ('IEEE', c_uint8 * 3),  # IEEE OUI Identifier
        ('TCIDA1LB', c_uint16),  # Telemetry Controller-Initiated Data Area 1 Last Block
        ('TCIDA2LB', c_uint16),  # Telemetry Controller-Initiated Data Area 2 Last Block
        ('TCIDA3LB', c_uint16),  # Telemetry Controller-Initiated Data Area 3 Last Block
        ('RSV1', c_uint8 * 368),  # Reserved 1
        ('TCIDA', c_uint8),  # Telemetry Controller-Initiated Data Available
        ('TCIDGN', c_uint8),  # Telemetry Controller-Initiated Data Generation Number
        ('RID', c_uint8 * 128),  # Reason Identifier
        ('TCIDB', TelemetryDataBlock * 7),  # Telemetry Host-Initiated Data Block 1 - n
    ]
    desc = {
        'LID': 'Log Identifier', 'IEEE': 'IEEE OUI Identifier',
        'TCIDA1LB': 'Telemetry Controller-Initiated Data Area 1 Last Block',
        'TCIDA2LB': 'Telemetry Controller-Initiated Data Area 2 Last Block',
        'TCIDA3LB': 'Telemetry Controller-Initiated Data Area 3 Last Block',
        'TCIDA': 'Telemetry Controller-Initiated Data Available',
        'TCIDGN': 'Telemetry Controller-Initiated Data Generation Number',
        'RID': 'Reason Identifier', 'TCIDB': 'Telemetry Controller-Initiated Data Block',
    }


class ReservationNotificationLog(Structure):
    """Get Log Page - Reservation Notification (Log Identifier 80h)"""
    _pack_ = 1
    _fields_ = [
        ('LPC', c_uint64),
        ('RNLPT', c_uint8),
        ('NALP', c_uint8),
        ('RSV0', c_uint8 * 2),
        ('NSID', c_uint32),
        ('RSV1', c_uint8 * 48),
    ]
    desc = {
        'LPC': 'Log Page Count', 'RNLPT': 'Reservation Notification Log Page Type',
        'NALP': 'Number of Available Log Pages', 'NSID': 'Namespace ID:',
    }


class SanitizeStatusLog(Structure):
    """Get Log Page - Reservation Notification (Log Identifier 81h)"""
    _pack_ = 1
    _fields_ = [
        ('SPROG', c_uint16),
        ('SSTAT', c_uint16),
        ('SCDW10', c_uint32),
        ('ETFO', c_uint32),
        ('ETFBE', c_uint32),
        ('ETFCE', c_uint32),
        ('ETFOWND', c_uint32),
        ('ETFBEWND', c_uint32),
        ('ETFCEWND', c_uint32),
        ('RAV0', c_uint8 * 480),
    ]
    desc = {
        'SPROG': 'Sanitize Progress', 'SSTAT': 'Sanitize Status',
        'SCDW10': 'Sanitize Command Dword 10 Information', 'ETFO': 'Estimated Time For Overwrite',
        'ETFBE': 'Estimated Time For Block Erase', 'ETFCE': 'Estimated Time For Crypto Erase:',
        'ETFOWND': 'Estimated Time For Overwrite With No-Deallocate Media Modification',
        'ETFBEWND': 'Estimated Time For Block Erase With No-Deallocate Media Modification',
        'ETFCEWND': 'Estimated Time For Crypto Erase With No-Deallocate Media Modification',
    }

class BadBlockHeader(Structure):
    """Get Log Page - CNEX Bad Block (Log Identifier CBh)"""
    _pack_ = 1
    _fields_ = [
        ('SN', Byte * 20),  # Serial Number
        ('RN', c_uint32),  # Revision Number
        ('NE', c_uint32),  # NAND Entities
        ('NOF', c_uint32),  # Number Of FBB
        ('NOM', c_uint32),  # Number Of MBB
        ('NOG', c_uint32),  # Number Of GBB
        ('FBS', c_uint32),  # FBB Bitmap Start
        ('MBS', c_uint32),  # MBB Bitmap Start
        ('GBS', c_uint32),  # GBB Bitmap Start
        ('BDS', c_uint32),  # BB Details Start
        ('TL', c_uint32),  # Total Length
        ('UUID', Uint128 * 128),  # UUIDs
        ('RSV', c_uint8 * 1988),
    ]
    desc = {
        'SN': 'Serial Number', 'RN': 'Revision Number', 'NE': 'NAND Entities', 'NOF': 'Number Of FBB',
        'NOM': 'Number Of MBB', 'NOG': 'Number Of GBB', 'FBS': 'FBB Bitmap Start', 'MBS': 'MBB Bitmap Start',
        'GBS': 'GBB Bitmap Start', 'BDS': 'BB Details Start', 'TL': 'Total Length',
    }


class BadBlockDetails(Union):
    """Get Log Page - CNEX Bad Block (Log Identifier CBh)"""

    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('TYPE', c_uint64, 3),  # 0: Erase Error, 1: Write Error, 2: Read Error
            ('POT', c_uint64, 13),  # Power On Time
            ('PC', c_uint64, 16),  # PE Cycle
            ('BL', c_uint64, 13),
            ('PL', c_uint64, 3),
            ('LN', c_uint64, 4),
            ('CH', c_uint64, 4),
            ('RSV', c_uint64, 8),
        ]
        desc = {
            'POT': 'Power On Time', 'PC': 'PE Cycle', 'BL': 'Block', 'PL': 'Plane',
            'LN': 'Lun', 'CH': 'Channel'
        }

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('QWORD', c_uint64),
        ('DWORD', c_uint32 * 2),
        ('BITS', BITS),
    ]
